from __future__ import absolute_import, division, print_function, unicode_literals
import time
import os
import json
import subprocess
import shutil
import string
import sys
import numpy as np
import ast
import glob
from builtins import range
from AdaptivePELE.constants import constants, blockNames
from AdaptivePELE.simulation import simulationTypes
from AdaptivePELE.atomset import atomset, RMSDCalculator
from AdaptivePELE.utilities import utilities
SKLEARN = True
try:
    from sklearn.cluster import KMeans
except ImportError:
    SKLEARN = False


class SimulationParameters:
    def __init__(self):
        self.processors = 0
        self.executable = ""
        self.templetizedControlFile = ""
        self.dataFolder = ""
        self.documentsFolder = ""
        self.iterations = 0
        self.peleSteps = 0
        self.seed = 0
        self.exitCondition = None
        self.boxCenter = None
        self.boxRadius = 20
        self.modeMovingBox = None
        self.runEquilibration = False
        self.equilibrationLength = None
        self.destination = None
        self.origin = None
        self.equilibrationMode = ""
        self.SASAforBox = None
        self.columnSASA = None
        self.reportName = None
        self.trajectoryName = None
        self.srun = False
        self.numberEquilibrationStructures = 10


class SimulationRunner:
    def __init__(self, parameters):
        self.parameters = parameters
        self.processorsToClusterMapping = []

    def runSimulation(self, runningControlFile=""):
        pass

    def hasExitCondition(self):
        """
            Check if an exit condition has been set

            :returns: bool -- True if an exit condition is set
        """
        return self.parameters.exitCondition is not None

    def checkExitCondition(self, clustering, outputFolder):
        """
            Check if the exit condition has been met

            :param clustering: Clustering object
            :type clustering: :py:class:`.Clustering`

            :returns: bool -- True if the exit condition is met
        """
        if self.parameters.exitCondition.type == simulationTypes.EXITCONDITION_TYPE.METRICMULTIPLETRAJS:
            return self.parameters.exitCondition.checkExitCondition(outputFolder)
        else:
            return self.parameters.exitCondition.checkExitCondition(clustering)

    def makeWorkingControlFile(self, workingControlFilename, dictionary, inputTemplate=None):
        """
            Substitute the values in the templetized control file

            :param workingControlFilename: Name of the template control file
            :type workingControlFilename: str
            :param dictionary: Dictonary containing the parameters to substitute
                in the control file
            :param inputFileTemplate: Template control file
            :type inputFileTemplate: str
            :type dictionary: dict
        """
        if inputTemplate is None:
            with open(self.parameters.templetizedControlFile, "r") as inputFile:
                inputFileContent = inputFile.read()
        else:
            inputFileContent = inputTemplate

        inputFileTemplate = string.Template(inputFileContent)
        outputFileContent = inputFileTemplate.substitute(dictionary)
        with open(workingControlFilename, "w") as outputFile:
            outputFileContent = outputFileContent.replace("'", '"')
            outputFile.write(outputFileContent)

    def updateMappingProcessors(self, mapping):
        """
            Update the value of the processorsToClusterMapping, a list with the
            snapshot from which the trajectories will start in the next iteration

            :param mapping: List with the snapshot from which the trajectories
                will start in the next iteration
            :type mapping: list
        """
        self.processorsToClusterMapping = mapping[1:]+[mapping[0]]

    def writeMappingToDisk(self, epochDir):
        """
            Write the processorsToClusterMapping to disk

            :param epochDir: Name of the folder where to write the
                processorsToClusterMapping
            :type epochDir: str
        """
        if len(self.processorsToClusterMapping) == 0:
            return
        with open(epochDir+"/processorMapping.txt", "w") as f:
            f.write(':'.join(map(str, self.processorsToClusterMapping)))

    def readMappingFromDisk(self, epochDir):
        """
            Read the processorsToClusterMapping from disk

            :param epochDir: Name of the folder where to write the
                processorsToClusterMapping
            :type epochDir: str
        """
        try:
            with open(epochDir+"/processorMapping.txt") as f:
                self.processorsToClusterMapping = map(ast.literal_eval, f.read().rstrip().split(':'))
        except IOError:
            sys.stderr.write("WARNING: processorMapping.txt not found, you might not be able to recronstruct fine-grained pathways\n")

    def setZeroMapping(self):
        """
            Set the processorsToClusterMapping to zero
        """
        self.processorsToClusterMapping = [0 for _ in range(1, self.parameters.processors)]

    def createMultipleComplexesFilenames(self, numberOfSnapshots, tmpInitialStructuresTemplate, iteration, equilibration=False):
        """
            Creates the string to substitute the complexes in the PELE control file

            :param numberOfSnapshots: Number of complexes to write
            :type numberOfSnapshots: int
            :param tmpInitialStructuresTemplate: Template with the name of the initial strutctures
            :type tmpInitialStructuresTemplate: str
            :param iteration: Epoch number
            :type iteration: int
            :param equilibration: Flag to mark wether the complexes are part of an
                equilibration run
            :type equilibration: bool

            :returns: str -- jsonString to be substituted in PELE control file
        """
        pass


class PeleSimulation(SimulationRunner):
    def __init__(self, parameters):
        SimulationRunner.__init__(self, parameters)
        self.type = simulationTypes.SIMULATION_TYPE.PELE

    def createSymbolicLinks(self):
        """
            Create symbolic links to Data and Documents folder if they don't exist
        """
        if not os.path.islink("Data"):
            os.system("ln -s " + self.parameters.dataFolder + " Data")
        if not os.path.islink("Documents"):
            os.system("ln -s " + self.parameters.documentsFolder + " Documents")

    def getNextIterationBox(self, clusteringObject, outputFolder, resname):
        """
            Select the box for the next epoch, currently selecting the COM of
            the cluster with max SASA

            :param clusteringObject: Clustering object
            :type clusteringObject: :py:class:`.Clustering`
            :param outputFolder: Folder to the trajectories
            :type outputFolder: str
            :param resname: Name of the ligand in the pdb
            :type resname: str

            :returns str: -- string to be substitued in PELE control file
        """
        metrics = utilities.getMetricsFromReportsInEpoch(self.parameters.reportName, outputFolder, self.parameters.processors)
        if self.parameters.modeMovingBox.lower() == blockNames.SimulationParams.modeMovingBoxBinding:
            SASAcluster = np.argmin(metrics[:, self.parameters.columnSASA])
            if metrics[SASAcluster, self.parameters.columnSASA] < self.parameters.SASAforBox:
                self.parameters.SASAforBox = metrics[SASAcluster, self.parameters.columnSASA]
            else:
                return
        elif self.parameters.modeMovingBox.lower() == blockNames.SimulationParams.modeMovingBoxUnBinding:
            SASAcluster = np.argmax(metrics[:, self.parameters.columnSASA])
            if metrics[SASAcluster, self.parameters.columnSASA] > self.parameters.SASAforBox:
                self.parameters.SASAforBox = metrics[SASAcluster, self.parameters.columnSASA]
            else:
                return
        else:
            raise ValueError("%s should be either binding or unbinding, but %s is provided!!!" % (blockNames.SimulationParams.modeMovingBox, self.parameters.modeMovingBox))
        # If this lines are reached then a new extreme SASA value was
        # identified and we proceed to extract the corresponding center of mass
        trajNum = metrics[SASAcluster, -2]
        snapshotNum = metrics[SASAcluster, -1]
        snapshot = utilities.getSnapshots(os.path.join(outputFolder, self.parameters.trajectoryName % trajNum))[int(snapshotNum)]
        snapshotPDB = atomset.PDB()
        snapshotPDB.initialise(snapshot, resname=resname)
        self.parameters.boxCenter = str(snapshotPDB.getCOM())
        return

    def selectInitialBoxCenter(self, initialStructuresAsString, resname):
        """
            Select the coordinates of the first box, currently as the center of
            mass of the first initial structure provided

            :param initialStructuresAsString: String containing the files of the initial structures
            :type initialStructuresAsString: str
            :param resname: Residue name of the ligand in the system pdb
            :type resname: str

            :returns str: -- string to be substitued in PELE control file
        """
        # This is a dictionary because it's prepared to be subtitued in the PELE
        # control files (i.e. JSON format)
        try:
            initialStructuresDict = json.loads(initialStructuresAsString.split(",")[0])
            initialStruct = str(initialStructuresDict['files'][0]['path'])
        except ValueError:
            # If a json valid string is not passed, interpret the input string
            # as a path to a file
            initialStruct = initialStructuresAsString

        PDBinitial = atomset.PDB()
        PDBinitial.initialise(initialStruct, resname=resname)
        return repr(PDBinitial.getCOM())

    def runSimulation(self, runningControlFile=""):
        """
            Run a short PELE simulation

            :param runningControlFile: PELE control file to run
            :type runningControlFile: str
        """
        self.createSymbolicLinks()

        if self.parameters.srun:
            toRun = ["srun", self.parameters.executable, runningControlFile]
        else:
            toRun = ["mpirun -np " + str(self.parameters.processors), self.parameters.executable, runningControlFile]
        toRun = " ".join(toRun)
        print(toRun)
        startTime = time.time()
        proc = subprocess.Popen(toRun, stdout=subprocess.PIPE, shell=True, universal_newlines=True)
        (out, err) = proc.communicate()
        print(out)
        if err:
            print(err)

        endTime = time.time()
        print("PELE took %.2f sec" % (endTime - startTime))

    def getEquilibrationControlFile(self, peleControlFileDict):
        """
            Filter unnecessary parameters and return a minimal PELE control
            file for equilibration runs

            :param peleControlFileDict: Dictionary with pele control file options
            :type peleControlFileDict: dict

            :returns: dict -- Dictionary with pele control file options
        """
        # Set small rotations and translations
        peleControlFileDict["commands"][0]["Perturbation"]["parameters"]["translationRange"] = 0.5
        peleControlFileDict["commands"][0]["Perturbation"]["parameters"]["rotationScalingFactor"] = 0.05
        # Remove dynamical changes in control file
        peleControlFileDict["commands"][0]["PeleTasks"][0].pop("exitConditions", None)
        peleControlFileDict["commands"][0]["PeleTasks"][0].pop("parametersChanges", None)
        # Set box_radius to 2
        peleControlFileDict["commands"][0]["Perturbation"]["Box"]["fixedCenter"] = "$BOX_CENTER"
        peleControlFileDict["commands"][0]["Perturbation"]["Box"]["radius"] = 2
        # Ensure random tags exists in metrics
        metricsBlock = peleControlFileDict["commands"][0]["PeleTasks"][0]["metrics"]
        nMetrics = len(metricsBlock)
        randomIndexes = sorted([i for i in range(nMetrics) if metricsBlock[i]['type'] == "random"])
        # Delete random numbers from metrics
        for index in randomIndexes[::-1]:
            del metricsBlock[index]
        nMetrics = len(metricsBlock)
        # Add new random number to metrics
        metricsBlock.extend([{'tag': 'rand', 'type': 'random'}])
        # Add equilibration dynamical changes
        changes = {
            "doThesechanges": {
                "Perturbation::parameters": {"rotationScalingFactor": 0.050}
            },
            "ifAnyIsTrue": ["rand >= .5"],
            "otherwise": {
                "Perturbation::parameters": {"rotationScalingFactor": 0.15}
            }
            }
        peleControlFileDict["commands"][0]["PeleTasks"][0]["parametersChanges"] = [changes]

        return peleControlFileDict

    def getMetricColumns(self, JSONdict):
        """
            Extract the column of a similarity distance (RMSD or distance) from the pele control file

            :param JSONdict: Dictionary containing a parsed PELE control file
            :type JSONdict: dict

            :returns: int -- Column index of the similarity metric
        """
        hasRMSD = False
        RMSDCol = None
        distanceCol = None
        hasDistance = False
        for i, metricBlock in enumerate(JSONdict["commands"][0]["PeleTasks"][0]['metrics']):
            if 'rmsd' in metricBlock['type'].lower():
                hasRMSD = True
                RMSDCol = i+4
            elif 'distance' in metricBlock['type'].lower():
                hasDistance = True
                distanceCol = i+4
        if hasRMSD:
            return RMSDCol
        elif hasDistance:
            return distanceCol
        else:
            return None

    def calculateEquilibrationLength(self):
        """
            Calculate the number of steps for the equilibration lenght accoding
            to the available processors
        """
        # Total steps is an approximate number of total steps to produce
        totalSteps = 1000
        # Take at least 5 steps
        stepsPerProc = max(int(totalSteps/float(self.parameters.processors)), 5)
        # but no more than 50
        return min(stepsPerProc, 50)

    def equilibrate(self, initialStructures, outputPathConstants, reportFilename, outputPath, resname):
        """
            Run short simulation to equilibrate the system. It will run one
            such simulation for every initial structure and select appropiate
            structures to start the simulation

            :param initialStructures: Name of the initial structures to copy
            :type initialStructures: list of str
            :param outputPathConstants: Contains outputPath-related constants
            :type outputPathConstants: :py:class:`.OutputPathConstants`
            :param reportBaseFilename: Name of the file that contains the metrics of the snapshots to cluster
            :type reportBaseFilename: str
            :param outputPath: Path where trajectories are found
            :type outputPath: str
            :param resname: Residue name of the ligand in the system pdb
            :type resname: str

            :returns: list --  List with initial structures
        """
        newInitialStructures = []
        if self.parameters.equilibrationLength is None:
            self.parameters.equilibrationLength = self.calculateEquilibrationLength()
        equilibrationPeleDict = {"PELE_STEPS": self.parameters.equilibrationLength, "SEED": self.parameters.seed}
        peleControlFileDict, templateNames = utilities.getPELEControlFileDict(self.parameters.templetizedControlFile)
        peleControlFileDict = self.getEquilibrationControlFile(peleControlFileDict)
        similarityColumn = self.getMetricColumns(peleControlFileDict)
        reportWildcard, trajWildcard = utilities.getReportAndTrajectoryWildcard(peleControlFileDict)

        for i, structure in enumerate(initialStructures):
            equilibrationOutput = os.path.join(outputPath, "equilibration_%d" % (i+1))
            equilibrationControlFile = outputPathConstants.tmpControlFilenameEqulibration % (i+1)
            utilities.makeFolder(equilibrationOutput)
            shutil.copyfile(structure, outputPathConstants.tmpInitialStructuresEquilibrationTemplate % (i+1))
            initialStructureString = self.createMultipleComplexesFilenames(1, outputPathConstants.tmpInitialStructuresEquilibrationTemplate, i+1, equilibration=True)
            equilibrationPeleDict["OUTPUT_PATH"] = equilibrationOutput
            equilibrationPeleDict["COMPLEXES"] = initialStructureString
            equilibrationPeleDict["BOX_CENTER"] = self.selectInitialBoxCenter(structure, resname)
            equilibrationPeleDict["BOX_RADIUS"] = 2
            for name in ["BOX_CENTER", "BOX_RADIUS"]:
                # If the template PELE control file is not templetized with the
                # box information, include it manually
                if name not in templateNames:
                    templateNames[name] = '"$%s"' % name
            print("Running equilibration for initial structure number %d" % (i+1))
            peleControlString = json.dumps(peleControlFileDict, indent=4)
            for key, value in templateNames.items():
                # Remove double quote around template keys, so that PELE
                # understands the options
                peleControlString = peleControlString.replace(value, "$%s" % key)

            self.makeWorkingControlFile(equilibrationControlFile, equilibrationPeleDict, peleControlString)
            self.runSimulation(equilibrationControlFile)
            # Extract report, trajnames, metrics columns from pele control file
            reportNames = os.path.join(equilibrationOutput, reportWildcard)
            trajNames = os.path.join(equilibrationOutput, trajWildcard)
            if len(initialStructures) == 1 and self.parameters.equilibrationMode == blockNames.SimulationParams.equilibrationLastSnapshot:
                newStructure = self.selectEquilibrationLastSnapshot(self.parameters.processors, trajNames)
            elif self.parameters.equilibrationMode == blockNames.SimulationParams.equilibrationSelect:
                newStructure = self.selectEquilibratedStructure(self.parameters.processors, similarityColumn, resname, trajNames, reportNames)
            elif self.parameters.equilibrationMode == blockNames.SimulationParams.equilibrationCluster:
                newStructure = self.clusterEquilibrationStructures(resname, trajNames, reportNames)

            for j, struct in enumerate(newStructure):
                newStructurePath = os.path.join(equilibrationOutput, 'equilibration_struc_%d_%d.pdb' % (i+1, j+1))
                with open(newStructurePath, "w") as fw:
                    fw.write(struct)
                newInitialStructures.append(newStructurePath)
        return newInitialStructures

    def clusterEquilibrationStructures(self, resname, trajWildcard, reportWildcard):
        """
            Cluster the equilibration run

            :param resname: Name of the ligand in the pdb
            :type resname: str
            :param trajWildcard: Templetized path to trajectory files"
            :type trajWildcard: str
            :param reportWildcard: Templetized path to report files"
            :type reportWildcard: str

            :returns: list -- List with the pdb snapshots of the representatives structures
        """
        if not SKLEARN:
            raise utilities.UnsatisfiedDependencyException("No installation of scikit-learn found. Please, install scikit-learn or select a different equilibrationMode.")
        energyColumn = 3
        # detect number of trajectories available
        nTrajs = len(glob.glob(trajWildcard.rsplit("_", 1)[0]+"*"))+1
        data = []
        for i in range(1, nTrajs):
            report = np.loadtxt(reportWildcard % i)
            if len(report.shape) < 2:
                report = report[np.newaxis, :]
            snapshots = utilities.getSnapshots(trajWildcard % i)
            for nSnap, (line, snapshot) in enumerate(zip(report, snapshots)):
                conformation = atomset.PDB()
                conformation.initialise(snapshot, resname=resname)
                com = conformation.getCOM()
                data.append([line[energyColumn], i, nSnap]+com)
        data = np.array(data)
        data = data[data[:, 0].argsort()]
        nPoints = max(self.parameters.numberEquilibrationStructures, data.shape[0]//4)
        data = data[:nPoints]
        kmeans = KMeans(n_clusters=self.parameters.numberEquilibrationStructures).fit(data[:, 3:])
        print("Clustered equilibration output into %d clusters!" % self.parameters.numberEquilibrationStructures)
        clustersInfo = {x: {"structure": None, "minDist": 1e6} for x in range(self.parameters.numberEquilibrationStructures)}
        for conf, cluster in zip(data, kmeans.labels_):
            dist = np.linalg.norm(kmeans.cluster_centers_[cluster]-conf[3:])
            if dist < clustersInfo[cluster]["minDist"]:
                clustersInfo[cluster]["minDist"] = dist
                clustersInfo[cluster]["structure"] = tuple(conf[1:3].astype(int))
        initialStructures = []
        for cl in range(self.parameters.numberEquilibrationStructures):
            if clustersInfo[cl]["structure"] is None:
                # If a cluster has no structure assigned, skip it
                continue
            traj, snap = clustersInfo[cl]["structure"]
            initialStructures.append(utilities.getSnapshots(trajWildcard % traj)[snap])
        return initialStructures

    def selectEquilibrationLastSnapshot(self, nTrajs, trajWildcard):
        """
            Select the last snapshot of each trajectory as a  representative
            initial structure from the equilibration run

            :param nTrajs: Number of trajectories
            :type nTrajs: int
            :param trajWildcard: Templetized path to trajectory files"
            :type trajWildcard: str

            :returns: list -- List with the pdb snapshots of the representatives structures
        """
        return [utilities.getSnapshots(trajWildcard % ij)[-1] for ij in range(1, nTrajs)]

    def selectEquilibratedStructure(self, nTrajs, similarityColumn, resname, trajWildcard, reportWildcard):
        """
            Select a representative initial structure from the equilibration
            run

            :param nTrajs: Number of trajectories
            :type nTrajs: int
            :param similarityColumn: Column number of the similarity metric
                (RMSD or distance)
            :type similarityColumn: int
            :param resname: Name of the ligand in the pdb
            :type resname: str
            :param trajWildcard: Templetized path to trajectory files"
            :type trajWildcard: str
            :param reportWildcard: Templetized path to report files"
            :type reportWildcard: str

            :returns: list -- List with the pdb snapshots of the representatives structures
        """
        energyColumn = 3
        values = []
        indices = []
        rowIndex = 0
        if similarityColumn is None:
            RMSDCalc = RMSDCalculator.RMSDCalculator()
            initial = atomset.PDB()
        else:
            cols = sorted([energyColumn, similarityColumn])

        for i in range(1, nTrajs):
            indices.append(rowIndex)
            report = np.loadtxt(reportWildcard % i)
            if similarityColumn is None:
                snapshots = utilities.getSnapshots(trajWildcard % i)
                report_values = []
                if i == 1:
                    initial.initialise(snapshots.pop(0), resname=resname)
                    report_values.append([0, report[0, energyColumn]])
                    report = report[1:, :]
                for j, snap in enumerate(snapshots):
                    pdbConformation = atomset.PDB()
                    pdbConformation.initialise(snap, resname=resname)
                    report_values.append([RMSDCalc.computeRMSD(initial, pdbConformation), report[j, energyColumn]])
            else:
                report_values = report[:, cols]
            values.extend(report_values)
            rowIndex += len(report_values)

        values = np.array(values)
        if energyColumn > similarityColumn or similarityColumn is None:
            similarityColumn, energyColumn = list(range(2))
        else:
            energyColumn, similarityColumn = list(range(2))
        maxEnergy = values.max(axis=0)[energyColumn]
        # Substract the max value so all values will be negative (avoid sign problems)
        values[:, energyColumn] -= maxEnergy
        minEnergy = values.min(axis=0)[energyColumn]
        # Normalize to the minimum value
        values[:, energyColumn] /= minEnergy
        freq, bins = np.histogram(values[:, similarityColumn])
        freq = np.divide(freq, np.float(np.sum(freq)))
        # Get center of the bins
        centers = (bins[:-1]+bins[1:])/2.0
        for row in values:
            row[similarityColumn] = freq[np.argmin(np.abs(row[similarityColumn] - centers))]
        distance = values.sum(axis=1)
        indexSelected = np.argmax(distance)
        # Get trajectory and snapshot number from the index selected
        trajNum = len(indices)-1
        for i, index in enumerate(indices):
            if index == indexSelected:
                trajNum = i
                break
            elif indexSelected < index:
                trajNum = i-1
                break
        snapshotNum = indexSelected-indices[trajNum]
        # trajectories are 1-indexed
        trajNum += 1
        # return as list for compatibility with selectEquilibrationLastSnapshot
        return [utilities.getSnapshots(trajWildcard % trajNum)[snapshotNum]]

    def createMultipleComplexesFilenames(self, numberOfSnapshots, tmpInitialStructuresTemplate, iteration, equilibration=False):
        """
            Creates the string to substitute the complexes in the PELE control file

            :param numberOfSnapshots: Number of complexes to write
            :type numberOfSnapshots: int
            :param tmpInitialStructuresTemplate: Template with the name of the initial strutctures
            :type tmpInitialStructuresTemplate: str
            :param iteration: Epoch number
            :type iteration: int
            :param equilibration: Flag to mark wether the complexes are part of an
                equilibration run
            :type equilibration: bool

            :returns: str -- jsonString to be substituted in PELE control file
        """
        jsonString = ["\n"]
        for i in range(numberOfSnapshots-1):
            if equilibration:
                jsonString.append(constants.inputFileTemplate % (tmpInitialStructuresTemplate % (iteration)) + ",\n")
            else:
                jsonString.append(constants.inputFileTemplate % (tmpInitialStructuresTemplate % (iteration, i)) + ",\n")
        if equilibration:
            jsonString.append(constants.inputFileTemplate % (tmpInitialStructuresTemplate % (iteration)))
        else:
            jsonString.append(constants.inputFileTemplate % (tmpInitialStructuresTemplate % (iteration, numberOfSnapshots-1)))
        return "".join(jsonString)


class TestSimulation(SimulationRunner):
    """
        Class used for testing
    """
    def __init__(self, parameters):
        SimulationRunner.__init__(self, parameters)
        self.type = simulationTypes.SIMULATION_TYPE.TEST
        self.copied = False
        self.parameters = parameters

    def runSimulation(self, runningControlFile=""):
        """
            Copy file to test the rest of the AdaptivePELE procedure
        """
        if not self.copied:
            if os.path.exists(self.parameters.destination):
                shutil.rmtree(self.parameters.destination)
            shutil.copytree(self.parameters.origin, self.parameters.destination)
            self.copied = True

    def makeWorkingControlFile(self, workingControlFilename, dictionary):
        pass


class ClusteringExitCondition:
    def __init__(self, ntrajs):
        self.clusterNum = 0
        self.ntrajs = ntrajs
        self.type = simulationTypes.EXITCONDITION_TYPE.CLUSTERING

    def checkExitCondition(self, clustering):
        """
            Iterate over all unchecked cluster and check if the exit condtion
            is met

            :param clustering: Clustering object
            :type clustering: :py:class:`.Clustering`

            :returns: bool -- Returns True if the exit condition has been met
        """
        newClusterNum = clustering.getNumberClusters()
        clusterDiff = newClusterNum - self.clusterNum
        self.clusterNum = newClusterNum
        return clusterDiff < 0.1*self.ntrajs


class MetricExitCondition:
    def __init__(self, metricCol, metricValue, condition):
        self.metricCol = metricCol
        self.metricValue = metricValue
        self.type = simulationTypes.EXITCONDITION_TYPE.METRIC
        if condition == ">":
            self.condition = lambda x, y: x > y
        else:
            self.condition = lambda x, y: x < y

    def checkExitCondition(self, clustering):
        """
            Iterate over all clusters and check if the exit condtion
            is met

            :param clustering: Clustering object
            :type clustering: :py:class:`.Clustering`

            :returns: bool -- Returns True if the exit condition has been met
        """
        for cluster in clustering.clusters.clusters:
            metric = cluster.getMetricFromColumn(self.metricCol)

            if metric is not None and self.condition(metric, self.metricValue):
                return True
        return False


class MetricMultipleTrajsExitCondition:
    def __init__(self, metricCol, metricValue, condition, reportWildCard, numTrajs, nProcessors):
        self.metricCol = metricCol
        self.metricValue = metricValue
        self.type = simulationTypes.EXITCONDITION_TYPE.METRICMULTIPLETRAJS
        self.numTrajs = numTrajs
        self.nProcessors = nProcessors
        self.trajsFound = 0
        if condition == ">":
            self.condition = lambda x, y: x.max() > y
        else:
            self.condition = lambda x, y: x.min() < y
        self.report = reportWildCard

    def checkExitCondition(self, outputFolder):
        """
            Iterate over all reports and check if the exit condtion
            is met

            :param clustering: Clustering object
            :type clustering: :py:class:`.Clustering`

            :returns: bool -- Returns True if the exit condition has been met
        """
        for j in range(1, self.nProcessors):
            report = np.loadtxt(os.path.join(outputFolder, self.report % j))
            if len(report.shape) < 2:
                # If a report has only one line, add another axis
                report = report[None]
            if self.condition(report[:, self.metricCol], self.metricValue):
                self.trajsFound += 1
        return self.trajsFound >= self.numTrajs


class RunnerBuilder:
    def build(self, simulationRunnerBlock):
        """
            Build the selected  SimulationRunner object

            :param simulationRunnerBlock: Block of the control file
                corresponding to the simulation step
            :type simulationRunnerBlock: dict

            :returns: :py:class:`.SimulationRunner` -- SimulationRunner object
                selected
        """
        simulationType = simulationRunnerBlock[blockNames.SimulationType.type]
        paramsBlock = simulationRunnerBlock[blockNames.SimulationParams.params]
        params = SimulationParameters()
        if simulationType == blockNames.SimulationType.pele:
            params.processors = paramsBlock[blockNames.SimulationParams.processors]
            params.dataFolder = paramsBlock.get(blockNames.SimulationParams.dataFolder, constants.DATA_FOLDER)
            params.documentsFolder = paramsBlock.get(blockNames.SimulationParams.documentsFolder, constants.DOCUMENTS_FOLDER)
            params.executable = paramsBlock.get(blockNames.SimulationParams.executable, constants.PELE_EXECUTABLE)
            params.templetizedControlFile = paramsBlock[blockNames.SimulationParams.templetizedControlFile]
            params.iterations = paramsBlock[blockNames.SimulationParams.iterations]
            params.peleSteps = paramsBlock[blockNames.SimulationParams.peleSteps]
            params.seed = paramsBlock[blockNames.SimulationParams.seed]
            params.modeMovingBox = paramsBlock.get(blockNames.SimulationParams.modeMovingBox)
            if params.modeMovingBox is not None:
                if params.modeMovingBox.lower() == blockNames.SimulationParams.modeMovingBoxBinding:
                    params.SASAforBox = 1.0
                elif params.modeMovingBox.lower() == blockNames.SimulationParams.modeMovingBoxUnBinding:
                    params.SASAforBox = 0.0
                peleDict, _ = utilities.getPELEControlFileDict(params.templetizedControlFile)
                params.reportName, params.trajectoryName = utilities.getReportAndTrajectoryWildcard(peleDict)
                params.columnSASA = utilities.getSASAcolumnFromControlFile(peleDict)
            params.boxCenter = paramsBlock.get(blockNames.SimulationParams.boxCenter)
            params.boxRadius = paramsBlock.get(blockNames.SimulationParams.boxRadius, 20)
            params.runEquilibration = paramsBlock.get(blockNames.SimulationParams.runEquilibration, False)
            params.equilibrationMode = paramsBlock.get(blockNames.SimulationParams.equilibrationMode, blockNames.SimulationParams.equilibrationSelect)
            params.equilibrationLength = paramsBlock.get(blockNames.SimulationParams.equilibrationLength)
            params.numberEquilibrationStructures = paramsBlock.get(blockNames.SimulationParams.numberEquilibrationStructures, 10)
            params.srun = paramsBlock.get(blockNames.SimulationParams.srun, False)
            exitConditionBlock = paramsBlock.get(blockNames.SimulationParams.exitCondition, None)
            if exitConditionBlock:
                exitConditionBuilder = ExitConditionBuilder()
                params.exitCondition = exitConditionBuilder.build(exitConditionBlock, params.templetizedControlFile, params.processors)

            return PeleSimulation(params)
        elif simulationType == blockNames.SimulationType.md:
            pass
        elif simulationType == blockNames.SimulationType.test:
            params.processors = paramsBlock[blockNames.SimulationParams.processors]
            params.destination = paramsBlock[blockNames.SimulationParams.destination]
            params.origin = paramsBlock[blockNames.SimulationParams.origin]
            params.iterations = paramsBlock[blockNames.SimulationParams.iterations]
            params.peleSteps = paramsBlock[blockNames.SimulationParams.peleSteps]
            params.seed = paramsBlock[blockNames.SimulationParams.seed]
            return TestSimulation(params)
        else:
            sys.exit("Unknown simulation type! Choices are: " + str(simulationTypes.SIMULATION_TYPE_TO_STRING_DICTIONARY.values()))


class ExitConditionBuilder:
    def build(self, exitConditionBlock, templetizedControlFile, nProcessors):
        """
            Build the selected exit condition object

            :param exitConditionBlock: Block of the control file
                corresponding to the exit condition
            :type exitConditionBlock: dict

            :returns: :py:class:`.MetricExitCondition` -- MetricExitCondition object
                selected
        """
        exitConditionType = exitConditionBlock[blockNames.ExitConditionType.type]
        exitConditionParams = exitConditionBlock[blockNames.SimulationParams.params]
        if exitConditionType == blockNames.ExitConditionType.metric:
            # Start counting the columns by 1
            metricCol = exitConditionParams[blockNames.SimulationParams.metricCol]-1
            metricValue = exitConditionParams[blockNames.SimulationParams.exitValue]
            condition = exitConditionParams.get(blockNames.SimulationParams.condition, "<")
            if condition not in [">", "<"]:
                raise ValueError("In MetricExitCondition the parameter condition only accepts > or <, but %s was passed" % condition)
            return MetricExitCondition(metricCol, metricValue, condition)
        elif exitConditionType == blockNames.ExitConditionType.metricMultipleTrajs:
            # Start counting the columns by 1
            metricCol = exitConditionParams[blockNames.SimulationParams.metricCol]-1
            metricValue = exitConditionParams[blockNames.SimulationParams.exitValue]
            numTrajs = exitConditionParams[blockNames.SimulationParams.numTrajs]
            condition = exitConditionParams.get(blockNames.SimulationParams.condition, "<")
            if condition not in [">", "<"]:
                raise ValueError("In MetricMultipleTrajsExitCondition the parameter condition only accepts > or <, but %s was passed" % condition)
            peleControlFileDict, _ = utilities.getPELEControlFileDict(templetizedControlFile)
            reportWildCard, _ = utilities.getReportAndTrajectoryWildcard(peleControlFileDict)
            return MetricMultipleTrajsExitCondition(metricCol, metricValue, condition, reportWildCard, numTrajs, nProcessors)
        elif exitConditionType == blockNames.ExitConditionType.clustering:
            ntrajs = exitConditionParams[blockNames.SimulationParams.trajectories]
            return ClusteringExitCondition(ntrajs)
        else:
            sys.exit("Unknown exit condition type! Choices are: " + str(simulationTypes.EXITCONDITION_TYPE_TO_STRING_DICTIONARY.values()))



import argparse
import numpy as np
from AdaptivePELE.atomset import atomset
from AdaptivePELE.utilities import utilities
import matplotlib.pyplot as plt


def parseArguments():
    desc = "Calculate the aggregate RMSF for every residue of the protein\n"
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument("trajectory", type=str, nargs='+', help="Path to the trajectory or pdbs to analyse")
    parser.add_argument("-ref", default=None, help="Reference structure, if not specified use average of the trajectories")
    parser.add_argument("-nRes", type=int, default=10, help="Number of top RMSF residues to display")
    args = parser.parse_args()
    return args.trajectory, args.ref, args.nRes


def extractAvgPDB(trajs):
    nSnapshots = 0
    avgStruct = {}
    snapshotsTot = []
    for traj in trajs:
        snapshots = utilities.getSnapshots(traj)
        for snapshot in snapshots:
            nSnapshots += 1
            PDB = atomset.PDB()
            PDB.initialise(snapshot, type="PROTEIN")
            snapshotsTot.append(PDB)
            for atomID, atom in PDB.atoms.items():
                if atomID in avgStruct:
                    avgStruct[atomID] += (atom.getAtomCoords()-avgStruct[atomID])/nSnapshots
                else:
                    avgStruct[atomID] = atom.getAtomCoords()
    return avgStruct, snapshotsTot


def mapReference(ref, trajs):
    refPDB = atomset.PDB()
    refPDB.initialise(ref, type="PROTEIN")
    avgStruct = {atomID: refPDB.atoms[atomID].getAtomCoords() for atomID in refPDB.atoms}
    snapshotsTot = []
    for traj in trajs:
        snapshots = utilities.getSnapshots(traj)
        for snapshot in snapshots:
            PDB = atomset.PDB()
            PDB.initialise(snapshot, type="PROTEIN")
            snapshotsTot.append(PDB)

    return avgStruct, snapshotsTot

if __name__ == "__main__":
    trajs, ref, nResidues = parseArguments()
    if ref is None:
        avgPDB, totPDBs = extractAvgPDB(trajs)
    else:
        avgPDB, totPDBs = mapReference(ref, trajs)
    RMSF = {atom: 0.0 for atom in avgPDB}
    residueMapping = {}
    # TODO: Handle multiple chains and insertion residues in PDB
    for PDBobj in totPDBs:
        for atomID, atom in PDBobj.atoms.items():
            RMSF[atomID] += np.sum((atom.getAtomCoords()-avgPDB[atomID])**2)
    for atomID, atom in PDBobj.atoms.items():
        if atom.resnum not in residueMapping:
            residueMapping[atom.resnum] = {atomID}
        else:
            residueMapping[atom.resnum].add(atomID)
    RMSFresidue = {}
    for residue, atoms in residueMapping.items():
        RMSFresidue[residue] = sum([RMSF[atom] for atom in atoms])
        RMSFresidue[residue] /= len(atoms)
        RMSFresidue[residue] = np.sqrt(RMSFresidue[residue])

    print("Residue\tRMSF")
    for res in sorted(RMSFresidue, key=lambda x: RMSFresidue[x], reverse=True)[:nResidues]:
        print("%s\t%.4f" % (res, RMSFresidue[res]))

    plt.plot(list(RMSFresidue.keys()), RMSFresidue.values(), 'x')
    plt.xlabel("Residue number")
    plt.ylabel("RMSF")
    plt.savefig("RMSF-residue.png")
    plt.show()



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import range
import sys
import glob
import numpy as np
import os
import pickle
from six import reraise as raise_
from AdaptivePELE.constants import blockNames
import AdaptivePELE.atomset.atomset as atomset
from AdaptivePELE.utilities import utilities
from AdaptivePELE.atomset import SymmetryContactMapEvaluator as sym
from AdaptivePELE.atomset import RMSDCalculator
from AdaptivePELE.clustering import clusteringTypes
from AdaptivePELE.clustering import thresholdcalculator
from scipy import stats
import heapq
try:
    import networkx as nx
    NETWORK = True
except ImportError:
    NETWORK = False


class Clusters:
    def __init__(self):
        self.clusters = []

    def __getstate__(self):
        # Defining pickling interface to avoid problems when working with old
        # simulations if the properties of the clustering-related classes have
        # changed
        state = {"clusters": self.clusters}
        return state

    def __setstate__(self, state):
        # Restore instance attributes
        self.clusters = state['clusters']

    def __len__(self):
        return len(self.clusters)

    def addCluster(self, cluster):
        """
            Add a new cluster

            :param cluster:  Cluster object to insert
            :type cluster: :py:class:`.Cluster`
        """
        self.clusters.append(cluster)

    def insertCluster(self, index, cluster):
        """
            Insert a cluster in a specified index

            :param index: Positions at which insert the cluster
            :type index: int
            :param cluster:  Cluster object to insert
            :type cluster: :py:class:`.Cluster`
        """
        self.clusters.insert(index, cluster)

    def getNumberClusters(self):
        """
            Get the number of clusters contained

            :returns: int -- Number of clusters contained
        """
        return len(self.clusters)

    def getCluster(self, clusterNum):
        """
            Get the cluster at position clusterNum

            :param clusterNum: Index of the cluster to retrieve
            :type clusterNum: int
            :returns: :py:class:`.Cluster` -- Cluster at position clusterNum
        """
        return self.clusters[clusterNum]

    def printClusters(self, verbose=False):
        """
            Print clusters information

            :param verbose: Flag to control the verbosity of the code (default is False)
            :type verbose: bool
        """
        for i, cluster in enumerate(self.clusters):
            print("--------------")
            print("CLUSTER #%d" % i)
            print("--------------")
            cluster.printCluster(verbose)
            print("")

    def __getitem__(self, key):
        return self.clusters[key]

    def __setitem__(self, key, value):
        self.clusters[key] = value

    def __delitem__(self, key):
        del self.clusters[key]

    def __eq__(self, other):
        return self.clusters == other.clusters

    def __iter__(self):
        for cluster in self.clusters:
            yield cluster


class ConformationNetwork:
    """
        Object that contains the conformation network, a network with clusters as
        nodes and edges representing trantions between clusters. The network is
        stored using the networkx package[1]

        References
        ----------
        .. [1] Networkx python package https://networkx.github.io
    """
    def __init__(self):
        if NETWORK:
            self.network = nx.DiGraph()
        else:
            self.network = None

    def __getstate__(self):
        # Defining pickling interface to avoid problems when working with old
        # simulations if the properties of the clustering-related classes have
        # changed
        state = {"network": self.network}
        return state

    def __setstate__(self, state):
        # Restore instance attributes
        self.network = state['network']
        if NETWORK and int(nx.__version__.split(".")[0]) > 1 and 'adj' in self.network.__dict__:
            for attr in ['node', 'adj', 'graph', 'pred', 'succ']:
                self.network.__dict__["_"+attr] = self.network.__dict__.pop(attr)

    def add_node(self, node, **kwargs):
        """
            Add a node to the network (wrapper for networkx method)

            :param node: Name of the node
            :type node: int
            :param kwargs: Set or change attributes using key=value.
            :type kwargs: keyword arguments, optional
        """
        if not NETWORK:
            return
        self.network.add_node(node, attr_dict=kwargs)

    def add_edge(self, source, target):
        """
            Add an edge to the network (wrapper for networkx method)

            :param source: Name of the source node
            :type source: int
            :param target: Name of the target node
            :type target: int
        """
        if not NETWORK:
            return
        if self.network.has_edge(source, target):
            self.network[source][target]['transition'] += 1
        else:
            self.network.add_edge(source, target, transition=1)

    def writeConformationNetwork(self, path):
        """
            Write the conformational network to file to visualize it

            :param path: Path where to write the network
            :type path: str
        """
        if not NETWORK:
            sys.stderr.write("Package networkx not found! Could not write network\n")
            return
        nx.write_edgelist(self.network, path)

    def writeFDT(self, path):
        """
            Write the first discovery tree to file in edgelist format to
            visualize it

            :param path: Path where to write the network
            :type path: str
        """
        if not NETWORK:
            sys.stderr.write("Package networkx not found! Could not write network\n")
            return
        with open(path, "w") as fw:
            for node, data in self.network.nodes(data=True):
                if data['parent'] != 'root':
                    fw.write("%d\t%d\n" % (data['parent'], node))

    def createPathwayToCluster(self, clusterLeave):
        """
            Retrace the FDT from a specific cluster to the root where it was
            discovered

            :param clusterLeave: End point of the pathway to reconstruct
            :type clusterLeave: int
            :returns: list -- List of snapshots conforming a pathway
        """
        pathway = []
        nodeLabel = clusterLeave
        while nodeLabel != "root":
            pathway.append(nodeLabel)
            nodeLabel = self.network.node[nodeLabel]['parent']
        return pathway[::-1]


class AltStructures:
    """
        Helper class, each cluster will have an instance of AltStructures that
        will maintain a priority queue (pq) of alternative structures to spawn
        from encoded as tuples (priority, PDB).
    """
    def __init__(self):
        self.altStructPQ = []
        self.limitSize = 10
        self.index = -1

    def __getstate__(self):
        # Defining pickling interface to avoid problems when working with old
        # simulations if the properties of the clustering-related classes have
        # changed
        state = {"altStructPQ": self.altStructPQ, "limitSize": self.limitSize, "index": self.index}
        return state

    def __setstate__(self, state):
        # Restore instance attributes
        self.limitSize = state['limitSize']
        self.altStructPQ = [(el[0], i, el[-1]) for i, el in enumerate(state['altStructPQ'])]
        self.index = state.get('index', len(self.altStructPQ)-1)

    def altSpawnSelection(self, centerPair):
        """
            Select an alternative PDB from the cluster center to spawn from

            :param centerPair: Tuple with the population of the representative structure
                and the PDB of said structure
            :type centerPair: int, :py:class:`.PDB`
            :returns: :py:class:`.PDB`, tuple -- PDB of the strucutre selected to spawn and tuple
                consisting of (epoch, trajectory, snapshot)

        """
        subpopulations = [i[0] for i in self.altStructPQ]
        totalSubpopulation = sum(subpopulations)
        # Create a list of the population distributed between the cluster
        # center and the alternative structures
        weights = 1.0/np.array([centerPair[0]-totalSubpopulation]+subpopulations)
        weights /= weights.sum()
        # This function only works on numpy >= 1.7, on life we have 1.6
        # ind = np.random.choice(range(len(self.altStructPQ)), p=weights)
        # Add one to account for the cluster representative
        r = stats.rv_discrete(values=(range(self.sizePQ()+1), weights))
        ind = r.rvs(size=10)[0]
        # The first value of the distribution is always the cluster center
        if ind == 0:
            print("cluster center")
            return centerPair[1], None
        else:
            # pick an alternative structure from the priority queue
            # The first element corresponds to the cluster center
            ind -= 1
            print("alternative structure")
            return self.altStructPQ[ind][2].pdb, self.altStructPQ[ind][2].trajPosition

    def cleanPQ(self):
        """
            Ensure that the alternative structures priority queue has no more
            elements than the limit in order to ensure efficiency
        """
        if len(self.altStructPQ) < self.limitSize:
            return
        limit = len(self.altStructPQ)
        del self.altStructPQ[self.limitSize-limit:]

    def addStructure(self, PDB, threshold, resname, resnum, resChain, contactThreshold, similarityEvaluator, trajPosition):
        """
            Perform a subclustering, with sub-clusters of size threshold/2

            :param PDB: Structure to cluster
            :type PDB: :py:class:`.PDB`
            :param threshold: Size of the cluster
            :type threshold: float
            :param resname: String containing the three letter name of the ligand in the pdb
            :type resname: str
            :param resnum: Integer containing the residue number of the ligand in the pdb
            :type resnum: int
            :param resChain: String containing the chain name of the ligand in the pdb
            :type resChain: str
            :param contactThreshold: Distance at which to atoms are considered in contact
            :type contactThreshold: float
            :param similarityEvaluator: Object that determinates the similarity between two structures
            :type similarityEvaluator: :py:class:`.SimilarityEvaluator`
            :param trajPosition: Tuple of (epoch, trajectory, snapshot) that permit
                identifying the structure added
            :type trajPosition: int, int, int

        """
        i = 0
        for _, _, subCluster in self.altStructPQ:
            _, distance = similarityEvaluator.isElement(PDB, subCluster, resname, resnum, resChain, contactThreshold)
            if distance < subCluster.threshold/2.0:
                subCluster.addElement([])
                del self.altStructPQ[i]
                heapq.heappush(self.altStructPQ, (subCluster.elements, self.updateIndex(), subCluster))
                if len(self.altStructPQ) > 2*self.limitSize:
                    self.cleanPQ()
                return
            i += 1
        newCluster = Cluster(PDB, thresholdRadius=threshold, contactThreshold=contactThreshold, contactMap=similarityEvaluator.contactMap, trajPosition=trajPosition)
        heapq.heappush(self.altStructPQ, (1, self.updateIndex(), newCluster))
        if len(self.altStructPQ) > 2*self.limitSize:
            self.cleanPQ()

    def updateIndex(self):
        """
            Update the index which represents chronological order of entries in
            the priority queue

            :returns: int -- Index of the following element
        """
        self.index += 1
        return self.index

    def sizePQ(self):
        """
            Get the number of sub-clusters stored in the priority queue

            :returns: int -- Number of sub-clusters stored in the priority queue
        """
        return len(self.altStructPQ)


class Cluster:
    """
        A cluster contains a representative structure(pdb), the number of
        elements, its density, threshold, number of contacts,
        a contactMap(sometimes) and a metric
    """
    def __init__(self, pdb, thresholdRadius=None, contactMap=None,
                 contacts=None, metrics=None, metricCol=None, density=None,
                 contactThreshold=8, altSelection=False, trajPosition=None):
        """
            :param pdb: Pdb of the representative structure
            :type pdb: :py:class:`.PDB`
            :param thresholdRadius: Threshold of the cluster
            :type thresholdRadius: float
            :param contactMap:  The contact map of the ligand and the protein
            :type contactMap: numpy.Array
            :param contacts: Ratio of the number of alpha carbons in contact with the ligand
            :type contacts: float
            :param metrics: Array of the metrics corresponding to the cluster
            :type metrics: numpy.Array
            :param metricCol: Column of the prefered metric
            :type metricCol: int
            :param density: Density of the cluster
            :type density: float
            :param contactThreshold: Distance between two atoms to be considered in contact (default 8)
            :type contactThreshold: float
            :param altSelection: Flag that controls wether to use the alternative structures (default 8)
            :type altSelection: bool
            :param trajPosition: Tuple of (epoch, trajectory, snapshot) that permit
                identifying the structure added
            :type trajPosition: int, int, int

        """
        self.pdb = pdb
        self.altStructure = AltStructures()
        self.elements = 1
        self.threshold = thresholdRadius
        self.density = density
        self.contacts = contacts
        self.contactMap = contactMap
        if metrics is None:
            metrics = []
        self.metrics = metrics
        self.originalMetrics = metrics
        self.metricCol = metricCol
        self.contactThreshold = contactThreshold
        self.altSelection = altSelection
        self.trajPosition = trajPosition

        if self.threshold is None:
            self.threshold2 = None
        else:
            self.threshold2 = thresholdRadius*thresholdRadius

    def __getstate__(self):
        # Defining pickling interface to avoid problems when working with old
        # simulations if the properties of the clustering-related classes have
        # changed
        state = {"pdb": self.pdb, "altStructure": self.altStructure,
                 "elements": self.elements, "threshold": self.threshold,
                 "densitiy": self.density, "contacts": self.contacts,
                 "contactMap": self.contactMap, "metrics": self.metrics,
                 "metricCol": self.metricCol, "threshold2": self.threshold2,
                 "contactThreshold": self.contactThreshold,
                 "altSelection": self.altSelection,
                 "originalMetrics": self.originalMetrics,
                 "trajPosition": self.trajPosition}
        return state

    def __setstate__(self, state):
        # Restore instance attributes
        self.pdb = state['pdb']
        self.altStructure = state.get('altStructure', AltStructures())
        self.elements = state['elements']
        self.threshold = state.get('threshold')
        self.density = state.get('density')
        self.contacts = state.get('contacts')
        self.contactMap = state.get('contactMap')
        self.metrics = state.get('metrics', [])
        self.originalMetrics = state.get('originalMetrics', [])
        self.metricCol = state.get('metricCol')
        self.threshold2 = state.get('threshold2')
        self.contactThreshold = state.get('contactThreshold', 8)
        self.altSelection = state.get('altSelection', False)
        self.trajPosition = state.get('trajPosition')

    def __len__(self):
        return self.elements

    def getMetric(self):
        """
            Get the value of the prefered metric if present, otherwise return None

            :returns: float -- Value of the prefered metric
        """
        if len(self.metrics) and self.metricCol is not None:
            return self.metrics[self.metricCol]
        else:
            return None

    def getMetricFromColumn(self, numcol):
        """
            Get the value of the metric in column numcol if present, otherwise return None

            :param numcol: Column of the desired metric
            :type numcol: int
            :returns: float -- Value of the prefered metric
        """
        if len(self.metrics):
            return self.metrics[numcol]
        else:
            return None

    def addElement(self, metrics):
        """
            Add a new element to the cluster

            :param metrics: Array of metrics of the new structure
            :type metrics: numpy.Array
        """
        self.elements += 1
        if self.metrics is None:
            # Special case where cluster in created during clustering of
            # initial structures
            self.metrics = metrics
            return
        if self.originalMetrics is None:
            self.originalMetrics = metrics
        if len(metrics) and len(self.metrics):
            # Set all metrics to the minimum value
            self.metrics = np.minimum(self.metrics, metrics)

    def printCluster(self, verbose=False):
        """
            Print cluster information

            :param verbose: Flag to control the verbosity of the code (default is False)
            :type verbose: bool
        """
        if verbose:
            print(self.pdb.printAtoms())
        print("Elements: ", self.elements)
        print("Metrics: ", self.metrics)
        if self.threshold != 0:
            print("Radius threshold: ", self.threshold)
        print("Number of contacts: %.2f" % self.contacts)

    def __str__(self):
        return "Cluster: elements=%d, threshold=%.3f, contacts=%.3f, density=%.3f" % (self.elements, self.threshold, self.contacts, self.density or 0.000)

    def writePDB(self, path):
        """
            Write the pdb of the representative structure to file

            :param path: Filename of the file to write
            :type path: str
        """
        self.pdb.writePDB(str(path))

    def getContacts(self):
        """
            Get the contacts ratio of the cluster

            :returns: float -- contact ratio of the cluster
        """
        return self.contacts

    def writeSpawningStructure(self, path):
        """
            Write the pdb of the chosen structure to spawn

            :param path: Filename of the file to write
            :type path: str
            :returns int, int, int: Tuple of (epoch, trajectory, snapshot) that permit
                identifying the structure added
        """
        if not self.altSelection or self.altStructure.sizePQ() == 0:
            print("cluster center")
            self.pdb.writePDB(str(path))
            return self.trajPosition
        else:
            spawnStruct, trajPosition = self.altStructure.altSpawnSelection((self.elements, self.pdb))
            spawnStruct.writePDB(str(path))
            if trajPosition is None:
                trajPosition = self.trajPosition
            return trajPosition

    def __eq__(self, other):
        return self.pdb == other.pdb\
             and self.elements == other.elements\
             and self.threshold == other.threshold\
             and self.contacts == other.contacts\
             and np.allclose(self.metrics, other.metrics)


class ClusteringEvaluator:
    def __init__(self):
        self.contactMap = None
        self.contacts = None

    def cleanContactMap(self):
        """
            Clean the attributes to prepare for next iteration
        """
        self.contactMap = None
        self.contacts = None


class ContactsClusteringEvaluator(ClusteringEvaluator):
    def __init__(self, RMSDCalculator_object):
        """
            Helper object to carry out the RMSD clustering

            :param RMSDCalculator: object that calculates the RMSD between two
                conformations
            :type RMSDCalculator: :py:class:`.RMSDCalculator`
        """
        ClusteringEvaluator.__init__(self)
        self.RMSDCalculator = RMSDCalculator_object
        self.contacts = None
        # Only here for compatibility purpose
        self.contactMap = None

    def __getstate__(self):
        # Defining pickling interface to avoid problems when working with old
        # simulations if the properties of the clustering-related classes have
        # changed
        state = {"RMSDCalculator": self.RMSDCalculator,
                 "contacts": self.contacts, "contactMap": self.contactMap}
        return state

    def __setstate__(self, state):
        # Restore instance attributes
        self.RMSDCalculator = state.get('RMSDCalculator', RMSDCalculator.RMSDCalculator())
        self.contacts = state.get('contacts')
        self.contactMap = state.get('contactMap')

    def isElement(self, pdb, cluster, resname, resnum, resChain, contactThresholdDistance):
        """
            Evaluate wether a conformation is a member of a cluster

            :param pdb: Structure to compare
            :type pdb: :py:class:`.PDB`
            :param cluster: Cluster to compare
            :type cluster: :py:class:`.Cluster`
            :param resname: String containing the three letter name of the ligand in the pdb
            :type resname: str
            :param resnum: Integer containing the residue number of the ligand in the pdb
            :type resnum: int
            :param resChain: String containing the chain name of the ligand in the pdb
            :type resChain: str
            :param contactThreshold: Distance between two atoms to be considered in contact (default 8)
            :type contactThreshold: float
            :returns: bool, float -- Whether the structure belong to the cluster and the distance between them
        """
        dist = self.RMSDCalculator.computeRMSD(cluster.pdb, pdb)
        return dist < cluster.threshold, dist

    def checkAttributes(self, pdb, resname, resnum, resChain, contactThresholdDistance):
        """
            Check wether all attributes are set for this iteration

            :param pdb: Structure to compare
            :type pdb: :py:class:`.PDB`
            :param resname: String containing the three letter name of the ligand in the pdb
            :type resname: str
            :param resnum: Integer containing the residue number of the ligand in the pdb
            :type resnum: int
            :param resChain: String containing the chain name of the ligand in the pdb
            :type resChain: str
            :param contactThreshold: Distance between two atoms to be considered in contact (default 8)
            :type contactThreshold: float
        """
        if self.contacts is None:
            self.contacts = pdb.countContacts(resname, contactThresholdDistance, resnum, resChain)

    def getInnerLimit(self, cluster):
        """
            Return the threshold of the cluster

            :param cluster: Cluster to compare
            :type cluster: :py:class:`.Cluster`
            :returns: float -- Threshold of the cluster
        """
        return cluster.threshold2


class CMClusteringEvaluator(ClusteringEvaluator):
    limitSlope = {8: 6, 6: 15, 4: 60, 10: 3}
    limitMax = {8: 2, 6: 0.8, 4: 0.2, 10: 4}

    def __init__(self, similarityEvaluator, symmetryEvaluator):
        """
            Helper object to carry out the RMSD clustering

            :param similarityEvaluator: object that calculates the similarity
                between two contact maps
            :type similarityEvaluator: :py:class:`.CMSimilarityEvaluator`
            :param symmetryEvaluator: object to introduce the symmetry  in the
                contacts maps
            :type symmetryEvaluator: :py:class:`.SymmetryContactMapEvaluator`
        """
        ClusteringEvaluator.__init__(self)
        self.similarityEvaluator = similarityEvaluator
        self.symmetryEvaluator = symmetryEvaluator
        self.contacts = None
        self.contactMap = None

    def __getstate__(self):
        # Defining pickling interface to avoid problems when working with old
        # simulations if the properties of the clustering-related classes have
        # changed
        state = {"similarityEvaluator": self.similarityEvaluator,
                 "symmetryEvaluator": self.symmetryEvaluator,
                 "contacts": self.contacts, "contactMap": self.contactMap}
        return state

    def __setstate__(self, state):
        # Restore instance attributes
        self.similarityEvaluator = state.get('similarityEvaluator')
        self.symmetryEvaluator = state.get('symmetryEvaluator')
        self.contacts = state.get('contacts')
        self.contactMap = state.get('contactMap')

    def isElement(self, pdb, cluster, resname, resnum, resChain, contactThresholdDistance):
        """
            Evaluate wether a conformation is a member of a cluster

            :param pdb: Structure to compare
            :type pdb: :py:class:`.PDB`
            :param cluster: Cluster to compare
            :type cluster: :py:class:`.Cluster`
            :param resname: String containing the three letter name of the ligand in the pdb
            :type resname: str
            :param resnum: Integer containing the residue number of the ligand in the pdb
            :type resnum: int
            :param resChain: String containing the chain name of the ligand in the pdb
            :type resChain: str
            :param contactThreshold: Distance between two atoms to be considered in contact (default 8)
            :type contactThreshold: float
            :returns: bool, float -- Whether the structure belong to the cluster and the distance between them
        """
        if self.contactMap is None:
            self.contactMap, self.contacts = self.symmetryEvaluator.createContactMap(pdb, resname, contactThresholdDistance, resnum, resChain)
            # self.contactMap, foo = self.symmetryEvaluator.createContactMap(pdb, resname, contactThresholdDistance)
            # self.contacts = pdb.countContacts(resname, 8)  # contactThresholdDistance)
        distance = self.similarityEvaluator.isSimilarCluster(self.contactMap, cluster.contactMap, self.symmetryEvaluator)
        return distance < cluster.threshold, distance

    def checkAttributes(self, pdb, resname, resnum, resChain, contactThresholdDistance):
        """
            Check wether all attributes are set for this iteration

            :param pdb: Structure to compare
            :type pdb: :py:class:`.PDB`
            :param resname: String containing the three letter name of the ligand in the pdb
            :type resname: str
            :param resnum: Integer containing the residue number of the ligand in the pdb
            :type resnum: int
            :param resChain: String containing the chain name of the ligand in the pdb
            :type resChain: str
            :param contactThreshold: Distance between two atoms to be considered in contact (default 8)
            :type contactThreshold: float
        """
        if self.contactMap is None:
            self.contactMap, self.contacts = self.symmetryEvaluator.createContactMap(pdb, resname, contactThresholdDistance, resnum, resChain)
            # self.contactMap, foo = self.symmetryEvaluator.createContactMap(pdb, resname, contactThresholdDistance)
            # self.contacts = pdb.countContacts(resname, 8)  # contactThresholdDistance)

    def getInnerLimit(self, cluster):
        """
            Return the threshold of the cluster

            :param cluster: Cluster to compare
            :type cluster: :py:class:`.Cluster`
            :returns: float -- Threshold of the cluster
        """
        # if cluster.contacts > self.limitMax[cluster.contactThreshold]:
        #     return 4.0
        # else:
        #     return 16-self.limitSlope[cluster.contactThreshold]*cluster.contacts

        # if cluster.contacts > 2.0:
        #     return 4.0
        # elif cluster.contacts <= 0.5:
        #     return 25.0
        # else:
        #     return 25-14*(cluster.contacts-0.5)

        if cluster.contacts > 2.0:
            return 4.0
        elif cluster.contacts < 0.5:
            return 25.0
        else:
            return 16-8*(cluster.contacts-0.5)

        # if cluster.contacts > 1.0:
        #     return 4.0
        # elif cluster.contacts > 0.75:
        #     return 9.0
        # elif cluster.contacts > 0.5:
        #     return 16.0
        # else:
        #      return 25


class Clustering:
    def __init__(self, resname="", resnum=0, resChain="", reportBaseFilename=None,
                 columnOfReportFile=None, contactThresholdDistance=8,
                 altSelection=False):
        """
            Base class for clustering methods, it defines a cluster method that
            contacts and accumulative inherit and use

            :param resname: String containing the three letter name of the ligand in the pdb
            :type resname: str
            :param resnum: Integer containing the residue number of the ligand in the pdb
            :type resnum: int
            :param resChain: String containing the chain name of the ligand in the pdb
            :type resChain: str
            :param reportBaseFilename: Name of the file that contains the metrics of the snapshots to cluster
            :type reportBaseFilename: str
            :param columnOfReportFile: Column of the report file that contain the metric of interest
            :type columnOfReportFile: int
            :param contactThresholdDistance: Distance at wich a ligand atom and a protein atom are
                considered in contact(default 8)
            :type contactThresholdDistance: float
        """
        self.type = "BaseClass"

        self.clusters = Clusters()
        if reportBaseFilename:
            self.reportBaseFilename = reportBaseFilename + "_%d"
        else:
            self.reportBaseFilename = None
        self.resname = resname
        self.resnum = resnum
        self.resChain = resChain
        self.col = columnOfReportFile
        self.contactThresholdDistance = contactThresholdDistance
        self.symmetries = []
        self.altSelection = altSelection
        self.conformationNetwork = ConformationNetwork()
        self.epoch = -1

    def __getstate__(self):
        # Defining pickling interface to avoid problems when working with old
        # simulations if the properties of the clustering-related classes have
        # changed
        state = {"type": self.type, "clusters": self.clusters,
                 "reportBaseFilename": self.reportBaseFilename,
                 "resname": self.resname, "resnum": self.resnum,
                 "resChain": self.resChain, "col": self.col,
                 "epoch": self.epoch, "symmetries": self.symmetries,
                 "conformationNetwork": self.conformationNetwork,
                 "contactThresholdDistance": self.contactThresholdDistance,
                 "altSelection": self.altSelection}
        return state

    def __setstate__(self, state):
        # Restore instance attributes
        self.type = state['type']
        self.clusters = state['clusters']
        self.reportBaseFilename = state.get('reportBaseFilename')
        self.resname = state.get('resname', "")
        self.resnum = state.get('resnum', 0)
        self.resChain = state.get('resChain', "")
        self.col = state.get('col')
        self.contactThresholdDistance = state.get('contactThresholdDistance', 8)
        self.symmetries = state.get('symmetries', [])
        self.altSelection = state.get('altSelection', False)
        self.conformationNetwork = state.get('conformationNetwork', ConformationNetwork())
        self.epoch = state.get('epoch', -1)

    def __str__(self):
        return "Clustering: nClusters: %d" % len(self.clusters)

    def __len__(self):
        return len(self.clusters)

    def __iter__(self):
        for cluster in self.clusters:
            yield cluster

    def __getitem__(self, key):
        return self.clusters[key]

    def __setitem__(self, key, value):
        self.clusters[key] = value

    def __delitem__(self, key):
        del self.clusters[key]

    def setCol(self, col):
        """
            Set the column of the prefered column to col

            :param col: Column of the prefered column
            :type col: int
        """
        self.col = col

        for cluster in self.clusters.clusters:
            cluster.metricCol = col

    def getCluster(self, clusterNum):
        """
            Get the cluster at index clusterNum

            :returns: :py:class:`.Cluster` -- Cluster at clusterNum
        """
        return self.clusters.getCluster(clusterNum)

    def emptyClustering(self):
        """
            Delete previous results of clustering object
        """
        self.clusters = Clusters()
        self.conformationNetwork = ConformationNetwork()
        self.epoch = -1

    def clusterIterator(self):
        """
            Iterator over the clusters
        """
        # TODO: may be interesting to add some condition to filter, check
        # itertools module, its probably implemented
        for cluster in self.clusters.clusters:
            yield cluster

    def getNumberClusters(self):
        """
            Get the number of clusters

            :returns: int -- Number of clusters
        """
        return self.clusters.getNumberClusters()

    def __eq__(self, other):
        return self.clusters == other.clusters\
            and self.reportBaseFilename == other.reportBaseFilename\
            and self.resname == other.resname\
            and self.resnum == other.resnum\
            and self.resChain == other.resChain\
            and self.col == other.col

    def cluster(self, paths, ignoreFirstRow=False):
        """
            Cluster the snaptshots contained in the paths folder

            :param paths: List of folders with the snapshots
            :type paths: list
        """
        self.epoch += 1
        trajectories = getAllTrajectories(paths)
        for trajectory in trajectories:
            trajNum = utilities.getTrajNum(trajectory)
            # origCluster = processorsToClusterMapping[trajNum-1]
            origCluster = None
            snapshots = utilities.getSnapshots(trajectory, True)

            if self.reportBaseFilename:
                reportFilename = os.path.join(os.path.split(trajectory)[0],
                                              self.reportBaseFilename % trajNum)
                metrics = np.loadtxt(reportFilename, ndmin=2)

                for num, snapshot in enumerate(snapshots):
                    if ignoreFirstRow and num == 0:
                        continue
                    try:
                        origCluster = self.addSnapshotToCluster(trajNum, snapshot, origCluster, num, metrics[num], self.col)
                    except IndexError as e:
                        message = (" in trajectory %d. This is usually caused by a mismatch between report files and trajectory files"
                                   " which in turn is usually caused by some problem in writing the files, e.g. quota")

                        # raise a new exception of the same type, with the same
                        # traceback but with an added message
                        raise_(IndexError, (str(e) + message % trajNum), sys.exc_info()[2])
            else:
                for num, snapshot in enumerate(snapshots):
                    if ignoreFirstRow and num == 0:
                        continue
                    origCluster = self.addSnapshotToCluster(trajNum, snapshot, origCluster, num)
        for cluster in self.clusters.clusters:
            cluster.altStructure.cleanPQ()

    def writeOutput(self, outputPath, degeneracy, outputObject, writeAll):
        """
            Writes all the clustering information in outputPath

            :param outputPath: Folder that will contain all the clustering information
            :type outputPath: str
            :param degeneracy: Degeneracy of each cluster. It must be in the same order
                as in the self.clusters list
            :type degeneracy: list
            :param outputObject: Output name for the pickle object
            :type outputObject: str
            :param writeAll: Wether to write pdb files for all cluster in addition
                of the summary
            :type writeAll: bool
        """
        utilities.cleanup(outputPath)
        utilities.makeFolder(outputPath)

        summaryFilename = os.path.join(outputPath, "summary.txt")
        with open(summaryFilename, 'w') as summaryFile:
            summaryFile.write("#cluster size degeneracy contacts threshold density metric\n")

            for i, cluster in enumerate(self.clusters.clusters):
                if writeAll:
                    outputFilename = "cluster_%d.pdb" % i
                    outputFilename = os.path.join(outputPath, outputFilename)
                    cluster.writePDB(outputFilename)

                metric = cluster.getMetric()
                if metric is None:
                    metric = "-"
                else:
                    metric = "%.3f" % metric
                degeneracy_cluster = 0
                if degeneracy is not None:
                    # degeneracy will be None if null spawning is used
                    degeneracy_cluster = degeneracy[i]

                writeString = "%d %d %d %.2f %.4f %.1f %s\n" % (i, cluster.elements,
                                                                degeneracy_cluster,
                                                                cluster.contacts,
                                                                cluster.threshold,
                                                                cluster.density or 1.0,
                                                                metric)
                summaryFile.write(writeString)

        with open(outputObject, 'wb') as f:
            pickle.dump(self, f, 2)

    def addSnapshotToCluster(self, trajNum, snapshot, origCluster, snapshotNum, metrics=None, col=None):
        """
            Cluster a snapshot using the leader algorithm

            :param trajNum: Trajectory number
            :type trajNum: int
            :param snapshot: Snapshot to add
            :type snapshot: str
            :param origCluster: Cluster found in the previos snapshot
            :type origCluster: int
            :param snapshotNum: Number of snapshot in its trajectory
            :type snapshotNum: int
            :param metrics: Array with the metrics of the snapshot
            :type metrics: numpy.Array
            :param col: Column of the desired metrics
            :type col: int
            :returns: int -- Cluster to which the snapshot belongs
        """
        if metrics is None:
            metrics = []
        pdb = atomset.PDB()
        pdb.initialise(snapshot, resname=self.resname, resnum=self.resnum, chain=self.resChain)
        self.clusteringEvaluator.cleanContactMap()
        for clusterNum, cluster in enumerate(self.clusters.clusters):
            scd = atomset.computeSquaredCentroidDifference(cluster.pdb, pdb)
            if scd > self.clusteringEvaluator.getInnerLimit(cluster):
                continue

            isSimilar, dist = self.clusteringEvaluator.isElement(pdb, cluster,
                                                                 self.resname, self.resnum,
                                                                 self.resChain, self.contactThresholdDistance)
            if isSimilar:
                if dist > cluster.threshold/2.0:
                    cluster.altStructure.addStructure(pdb, cluster.threshold, self.resname, self.resnum, self.resChain, self.contactThresholdDistance, self.clusteringEvaluator, trajPosition=(self.epoch, trajNum, snapshotNum))
                cluster.addElement(metrics)
                if origCluster is None:
                    origCluster = clusterNum
                self.conformationNetwork.add_edge(origCluster, clusterNum)
                return clusterNum

        # if made it here, the snapshot was not added into any cluster
        # Check if contacts and contactMap are set (depending on which kind
        # of clustering)
        self.clusteringEvaluator.checkAttributes(pdb, self.resname, self.resnum,
                                                 self.resChain, self.contactThresholdDistance)
        contacts = self.clusteringEvaluator.contacts
        numberOfLigandAtoms = pdb.getNumberOfAtoms()
        contactsPerAtom = contacts/numberOfLigandAtoms

        threshold = self.thresholdCalculator.calculate(contactsPerAtom)
        cluster = Cluster(pdb, thresholdRadius=threshold,
                          contacts=contactsPerAtom,
                          contactMap=self.clusteringEvaluator.contactMap,
                          metrics=metrics, metricCol=col,
                          contactThreshold=self.contactThresholdDistance,
                          altSelection=self.altSelection, trajPosition=(self.epoch, trajNum, snapshotNum))
        self.clusters.addCluster(cluster)
        clusterNum = self.clusters.getNumberClusters()-1
        if clusterNum == origCluster or origCluster is None:
            origCluster = clusterNum
            # The clusterNum should only be equal to origCluster when the first
            # cluster is created and the clusterInitialStructures function has
            # not been called, i.e. when usind the compareClustering script
            self.conformationNetwork.add_node(clusterNum, parent='root', epoch=self.epoch)
        else:
            self.conformationNetwork.add_node(clusterNum, parent=origCluster, epoch=self.epoch)
        self.conformationNetwork.add_edge(origCluster, clusterNum)
        # If a new cluster is discovered during a trajectory, the next step in
        # the same trajectory will be considered to start from these new
        # cluster, thus resulting in a more precise conformation network and
        # smoother pathways
        return clusterNum

    def writeClusterMetric(self, path, metricCol):
        """
            Write the metric of each node in the conformation network in a
            tab-separated file

            :param path: Path where to write the network
            :type path: str
            :param metricCol: Column of the metric of interest
            :type metricCol: int
        """
        with open(path, "w") as f:
            for i, cluster in enumerate(self.clusters.clusters):
                metric = cluster.getMetricFromColumn(metricCol)
                if metric is None:
                    f.write("%d\t-\n" % i)
                else:
                    f.write("%d\t%.4f\n" % (i, metric))

    def writeConformationNodePopulation(self, path):
        """
            Write the population of each node in the conformation network in a
            tab-separated file

            :param path: Path where to write the network
            :type path: str
        """
        with open(path, "w") as f:
            for i, cluster in enumerate(self.clusters.clusters):
                f.write("%d\t%d\n" % (i, cluster.elements))

    def getOptimalMetric(self, column=None, simulationType="min"):
        """
            Find the cluster with the best metric

            :param column: Column of the metric that defines the best cluster,
                if not specified, the cluster metric is chosen
            :type column: int
            :param simulationType: Define optimal metric as the maximum or minimum, max or min
            :type simulationType: str

            :returns: int -- Number of cluster with the optimal metric
        """
        metrics = []
        for _, cluster in enumerate(self.clusters.clusters):
            if column is None:
                metric = cluster.getMetric()
            else:
                metric = cluster.getMetricFromColumn(column)
            metrics.append(metric)

        if simulationType.lower() == "min":
            optimalMetricIndex = np.argmin(metrics)
        elif simulationType.lower() == "max":
            optimalMetricIndex = np.argmax(metrics)
        else:
            raise ValueError("Unrecognized type simulation parameter!!! Possible values are max or min")

        return optimalMetricIndex

    def writePathwayTrajectory(self, pathway, filename):
        """
            Write a list of cluster forming a pathway into a trajectory pdb file

            :param pathway: List of clusters that form the pathway
            :type pathway: list
            :param filename: Path where to write the trajectory
            :type filename: str
        """
        with open(filename, "w") as pathwayFile:
            pathwayFile.write("REMARK 000 File created using PELE++\n")
            pathwayFile.write("REMARK 000 Pathway trajectory created using the FDT\n")
            pathwayFile.write("REMARK 000 List of cluster belonging to the pathway %s\n" % ' '.join(map(str, pathway)))
            for i, step_cluster in enumerate(pathway):
                cluster = self.clusters.clusters[step_cluster]
                pathwayFile.write("MODEL %d\n" % (i+1))
                pdbStr = cluster.pdb.pdb
                pdbList = pdbStr.split("\n")
                for line in pdbList:
                    line = line.strip()
                    # Avoid writing previous REMARK block
                    if line.startswith("REMARK ") or line.startswith("MODEL ") or line == "END":
                        continue
                    elif line:
                        pathwayFile.write(line+"\n")
                pathwayFile.write("ENDMDL\n")

    def writePathwayOptimalCluster(self, filename):
        """
            Extract the pathway to the cluster with the best metric as a
            trajectory and  write it to a PDB file

            :param filename: Path where to write the trajectory
            :type filename: str
        """
        optimalCluster = self.getOptimalMetric()
        pathway = self.createPathwayToCluster(optimalCluster)
        self.writePathwayTrajectory(pathway, filename)


class ContactsClustering(Clustering):
    def __init__(self, thresholdCalculator, resname="", resnum=0, resChain="",
                 reportBaseFilename=None, columnOfReportFile=None,
                 contactThresholdDistance=8, symmetries=None, altSelection=False):
        """
            Cluster together all snapshots that are closer to the cluster center
            than certain threshold. This threshold is assigned according to the
            ratio of number of contacts over the number of heavy atoms of the ligand

            :param resname: String containing the three letter name of the ligand in the pdb
            :type resname: str
            :param resnum: Integer containing the residue number of the ligand in the pdb
            :type resnum: int
            :param resChain: String containing the chain name of the ligand in the pdb
            :type resChain: str
            :param thresholdCalculator: ThresholdCalculator object that calculate the
                threshold according to the contacts ratio
            :type thresholdCalculator: :py:class:`.ThresholdCalculator`
            :param reportBaseFilename: Name of the file that contains the metrics of
                the snapshots to cluster
            :type reportBaseFilename: str
            :param columnOfReportFile: Column of the report file that contain the
                metric of interest
            :type columnOfReportFile: int
            :param contactThresholdDistance: Distance at wich a ligand atom and a protein atom are
                considered in contact(default 8)
            :type contactThresholdDistance: float
            :param symmetries: List of symmetric groups
            :type symmetries: list
            :param altSelection: Flag that controls wether to use the alternative structures (default 8)
            :type altSelection: bool
        """
        Clustering.__init__(self, resname=resname, resnum=resnum, resChain=resChain,
                            reportBaseFilename=reportBaseFilename,
                            columnOfReportFile=columnOfReportFile,
                            contactThresholdDistance=contactThresholdDistance,
                            altSelection=altSelection)
        self.type = clusteringTypes.CLUSTERING_TYPES.rmsd
        self.thresholdCalculator = thresholdCalculator
        if symmetries is None:
            symmetries = []
        self.symmetries = symmetries
        self.clusteringEvaluator = ContactsClusteringEvaluator(RMSDCalculator.RMSDCalculator(symmetries))

    def __getstate__(self):
        # Defining pickling interface to avoid problems when working with old
        # simulations if the properties of the clustering-related classes have
        # changed
        state = {"type": self.type, "clusters": self.clusters,
                 "reportBaseFilename": self.reportBaseFilename,
                 "resname": self.resname, "resnum": self.resnum,
                 "resChain": self.resChain, "col": self.col,
                 "epoch": self.epoch,
                 "symmetries": self.symmetries,
                 "conformationNetwork": self.conformationNetwork,
                 "contactThresholdDistance": self.contactThresholdDistance,
                 "altSelection": self.altSelection,
                 "thresholdCalculator": self.thresholdCalculator,
                 "clusteringEvaluator": self.clusteringEvaluator}
        return state

    def __setstate__(self, state):
        # Restore instance attributes
        self.type = state['type']
        self.clusters = state['clusters']
        self.reportBaseFilename = state.get('reportBaseFilename')
        self.resname = state.get('resname', "")
        self.resnum = state.get('resnum', 0)
        self.resChain = state.get('resChain', "")
        self.col = state.get('col')
        self.contactThresholdDistance = state.get('contactThresholdDistance', 8)
        self.symmetries = state.get('symmetries', [])
        self.altSelection = state.get('altSelection', False)
        self.conformationNetwork = state.get('conformationNetwork', ConformationNetwork())
        self.epoch = state.get('epoch', -1)
        self.thresholdCalculator = state.get('thresholdCalculator', thresholdcalculator.ThresholdCalculatorConstant())
        if isinstance(self.symmetries, dict):
            self.symmetries = [self.symmetries]
        self.clusteringEvaluator = state.get('clusteringEvaluator', ContactsClusteringEvaluator(RMSDCalculator.RMSDCalculator(self.symmetries)))


class ContactMapAccumulativeClustering(Clustering):
    def __init__(self, thresholdCalculator, similarityEvaluator, resname="",
                 resnum=0, resChain="",
                 reportBaseFilename=None, columnOfReportFile=None,
                 contactThresholdDistance=8, symmetries=None, altSelection=False):
        """ Cluster together all snapshots that have similar enough contactMaps.
            This similarity can be calculated with different methods (see similariyEvaluator documentation)

            :param thresholdCalculator: ThresholdCalculator object that calculate the
                threshold according to the contacts ratio
            :type thresholdCalculator: :py:class:`.ThresholdCalculator`
            :param similarityEvaluator: object that calculates the similarity
                between two contact maps
            :type similarityEvaluator: object
            :param resname: String containing the three letter name of the ligand in the pdb
            :type resname: str
            :param resnum: Integer containing the residue number of the ligand in the pdb
            :type resnum: int
            :param resChain: String containing the chain name of the ligand in the pdb
            :type resChain: str
            :param reportBaseFilename: Name of the file that contains the metrics of
                the snapshots to cluster
            :type reportBaseFilename: str
            :param columnOfReportFile: Column of the report file that contain the
                metric of interest
            :type columnOfReportFile: int
            :param contactThresholdDistance: Distance at wich a ligand atom and a protein atom are
                considered in contact(default 8)
            :type contactThresholdDistance: float
            :param symmetries: List of symmetric groups
            :type symmetries: list
            :param altSelection: Flag that controls wether to use the alternative structures (default 8)
            :type altSelection: bool
        """
        if symmetries is None:
            symmetries = []
        Clustering.__init__(self, resname=resname, resnum=resnum, resChain=resChain,
                            reportBaseFilename=reportBaseFilename,
                            columnOfReportFile=columnOfReportFile,
                            contactThresholdDistance=contactThresholdDistance,
                            altSelection=altSelection)
        self.type = clusteringTypes.CLUSTERING_TYPES.contactMap
        self.thresholdCalculator = thresholdCalculator
        self.similarityEvaluator = similarityEvaluator
        self.symmetryEvaluator = sym.SymmetryContactMapEvaluator(symmetries)
        self.clusteringEvaluator = CMClusteringEvaluator(similarityEvaluator, self.symmetryEvaluator)

    def __getstate__(self):
        # Defining pickling interface to avoid problems when working with old
        # simulations if the properties of the clustering-related classes have
        # changed
        state = {"type": self.type, "clusters": self.clusters,
                 "reportBaseFilename": self.reportBaseFilename,
                 "resname": self.resname, "resnum": self.resnum,
                 "resChain": self.resChain, "col": self.col,
                 "epoch": self.epoch, "symmetries": self.symmetries,
                 "conformationNetwork": self.conformationNetwork,
                 "contactThresholdDistance": self.contactThresholdDistance,
                 "altSelection": self.altSelection,
                 "thresholdCalculator": self.thresholdCalculator,
                 "similariyEvaluator": self.similarityEvaluator,
                 "symmetryEvaluator": self.symmetryEvaluator,
                 "clusteringEvaluator": self.clusteringEvaluator}
        return state

    def __setstate__(self, state):
        # Restore instance attributes
        self.type = state['type']
        self.clusters = state['clusters']
        self.reportBaseFilename = state.get('reportBaseFilename')
        self.resname = state.get('resname', "")
        self.resnum = state.get('resnum', 0)
        self.resChain = state.get('resChain', "")
        self.col = state.get('col')
        self.contactThresholdDistance = state.get('contactThresholdDistance', 8)
        self.symmetries = state.get('symmetries', [])
        self.altSelection = state.get('altSelection', False)
        self.conformationNetwork = state.get('conformationNetwork', ConformationNetwork())
        self.epoch = state.get('epoch', -1)
        self.thresholdCalculator = state.get('thresholdCalculator', thresholdcalculator.ThresholdCalculatorConstant(value=0.3))
        self.similarityEvaluator = state.get('similariyEvaluator', CMSimilarityEvaluator(blockNames.ClusteringTypes.Jaccard))
        self.symmetryEvaluator = state.get('symmetryEvaluator', sym.SymmetryContactMapEvaluator(self.symmetries))
        self.clusteringEvaluator = state.get('clusteringEvaluator', CMClusteringEvaluator(self.similarityEvaluator, self.symmetryEvaluator))


class SequentialLastSnapshotClustering(Clustering):
    """
        Assigned  the last snapshot of the trajectory to a cluster.
        Only useful for PELE sequential runs
    """
    def cluster(self, paths):
        """
            Cluster the snaptshots contained in the paths folder

            :param paths: List of folders with the snapshots
            :type paths: list
        """
        # Clean clusters at every step, so we only have the last snapshot of
        # each trajectory as clusters
        self.clusters = Clusters()
        trajectories = getAllTrajectories(paths)
        for trajectory in trajectories:
            trajNum = utilities.getTrajNum(trajectory)

            snapshots = utilities.getSnapshots(trajectory, True)
            if self.reportBaseFilename:
                reportFilename = os.path.join(os.path.split(trajectory)[0],
                                              self.reportBaseFilename % trajNum)
                metrics = np.loadtxt(reportFilename, ndmin=2)
                # Pass as cluster metrics the minimum value for each metric,
                # thus the metrics are not valid to do any spawning, only to
                # check the exit condition
                metrics = metrics.min(axis=0)

                self.addSnapshotToCluster(snapshots[-1], metrics, self.col)
            else:
                self.addSnapshotToCluster(snapshots[-1])

    def addSnapshotToCluster(self, snapshot, metrics=None, col=None):
        """
            Cluster a snapshot using the leader algorithm

            :param trajNum: Trajectory number
            :type trajNum: int
            :param snapshot: Snapshot to add
            :type snapshot: str
            :param metrics: Array with the metrics of the snapshot
            :type metrics: numpy.Array
            :param col: Column of the desired metrics
            :type col: int
            :returns: int -- Cluster to which the snapshot belongs
        """
        if metrics is None:
            metrics = []
        pdb = atomset.PDB()
        pdb.initialise(snapshot, resname=self.resname, resnum=self.resnum,
                       chain=self.resChain)
        contacts = pdb.countContacts(self.resname, self.contactThresholdDistance, self.resnum, self.resChain)
        numberOfLigandAtoms = pdb.getNumberOfAtoms()
        contactsPerAtom = contacts/numberOfLigandAtoms

        cluster = Cluster(pdb, thresholdRadius=0,
                          contacts=contactsPerAtom, metrics=metrics,
                          metricCol=col)
        self.clusters.addCluster(cluster)


class ClusteringBuilder:
    def buildClustering(self, clusteringBlock, reportBaseFilename=None, columnOfReportFile=None):
        """
            Builder to create the appropiate clustering object

            :param clusteringBlock: Parameters of the clustering process
            :type clusteringBlock: dict
            :param reportBaseFilename: Name of the file that contains the metrics of
                the snapshots to cluster
            :type reportBaseFilename: str
            :param columnOfReportFile: Column of the report file that contain the
                metric of interest
            :type columnOfReportFile: int
            :returns: :py:class:`.Clustering` -- Clustering object selected
        """
        paramsBlock = clusteringBlock[blockNames.ClusteringTypes.params]
        try:
            clusteringType = clusteringBlock[blockNames.ClusteringTypes.type]
            contactThresholdDistance = paramsBlock.get(blockNames.ClusteringTypes.contactThresholdDistance, 8)
            altSelection = paramsBlock.get(blockNames.ClusteringTypes.alternativeStructure, False)
        except KeyError as err:
            err.message += ": Need to provide mandatory parameter in clustering block"
            raise KeyError(err.message)
        resname = str(paramsBlock.get(blockNames.ClusteringTypes.ligandResname, "")).upper()
        resnum = int(paramsBlock.get(blockNames.ClusteringTypes.ligandResnum, 0))
        resChain = str(paramsBlock.get(blockNames.ClusteringTypes.ligandChain, "")).upper()
        if clusteringType == blockNames.ClusteringTypes.rmsd:
            symmetries = paramsBlock.get(blockNames.ClusteringTypes.symmetries, [])

            thresholdCalculatorBuilder = thresholdcalculator.ThresholdCalculatorBuilder()
            thresholdCalculator = thresholdCalculatorBuilder.build(clusteringBlock)
            return ContactsClustering(thresholdCalculator, resname=resname, resnum=resnum, resChain=resChain,
                                      reportBaseFilename=reportBaseFilename, columnOfReportFile=columnOfReportFile,
                                      contactThresholdDistance=contactThresholdDistance, symmetries=symmetries,
                                      altSelection=altSelection)
        elif clusteringType == blockNames.ClusteringTypes.lastSnapshot:

            return SequentialLastSnapshotClustering(resname=resname, resnum=resnum, resChain=resChain,
                                                    reportBaseFilename=reportBaseFilename,
                                                    columnOfReportFile=columnOfReportFile,
                                                    contactThresholdDistance=contactThresholdDistance)
        elif clusteringType == blockNames.ClusteringTypes.contactMap:
            symmetries = paramsBlock.get(blockNames.ClusteringTypes.symmetries, [])
            thresholdCalculatorBuilder = thresholdcalculator.ThresholdCalculatorBuilder()
            thresholdCalculator = thresholdCalculatorBuilder.build(clusteringBlock)
            try:
                similarityEvaluatorType = paramsBlock[blockNames.ClusteringTypes.similarityEvaluator]
            except KeyError:
                raise ValueError("No similarity Evaluator specified!!")
            similarityBuilder = similarityEvaluatorBuilder()
            similarityEvaluator = similarityBuilder.build(similarityEvaluatorType)
            return ContactMapAccumulativeClustering(thresholdCalculator, similarityEvaluator, resname=resname,
                                                    resnum=resnum, resChain=resChain,
                                                    reportBaseFilename=reportBaseFilename, columnOfReportFile=columnOfReportFile,
                                                    contactThresholdDistance=contactThresholdDistance, symmetries=symmetries, altSelection=altSelection)
        else:
            sys.exit("Unknown clustering method! Choices are: " +
                     str(clusteringTypes.CLUSTERING_TYPE_TO_STRING_DICTIONARY.values()))


class similarityEvaluatorBuilder:
    def build(self, similarityEvaluatorType):
        """
            Builder to create the appropiate similarityEvaluator

            :param similarityEvaluatorType: Type of similarityEvaluator chosen
            :type similarityEvaluatorType: str
            :returns: :py:class:`.SimilarityEvaluator` -- SimilarityEvaluator object selected
        """
        if similarityEvaluatorType in clusteringTypes.SIMILARITY_TYPES_NAMES:
            return CMSimilarityEvaluator(similarityEvaluatorType)
        else:
            sys.exit("Unknown threshold calculator type! Choices are: " + str(clusteringTypes.SIMILARITY_TYPES_TO_STRING_DICTIONARY.values()))


class CMSimilarityEvaluator:
    """
        Evaluate the similarity of two contactMaps by calculating the ratio of
        the number of differences over the average of elements in the contacts
        maps, their correlation or their Jaccard index, that is, the ratio
        between the intersection of the two contact maps and their union
    """
    def __init__(self, typeEvaluator):
        self.typeEvaluator = typeEvaluator

    def isSimilarCluster(self, contactMap, clusterContactMap, symContactMapEvaluator):
        """
            Evaluate if two contactMaps are similar or not, return True if yes,
            False otherwise

            :param contactMap: contactMap of the structure to compare
            :type contactMap: numpy.Array
            :param contactMap: contactMap of the structure to compare
            :type contactMap: numpy.Array
            :param symContactMapEvaluator: Contact Map symmetry evaluator object
            :type symContactMapEvaluator: :py:class:`.SymmetryContactMapEvaluator`
            :returns: float -- distance between contact maps
        """
        if self.typeEvaluator == blockNames.ClusteringTypes.correlation:
            return symContactMapEvaluator.evaluateCorrelation(contactMap, clusterContactMap)
        elif self.typeEvaluator == blockNames.ClusteringTypes.Jaccard:
            return symContactMapEvaluator.evaluateJaccard(contactMap, clusterContactMap)
        elif self.typeEvaluator == blockNames.ClusteringTypes.differenceDistance:
            return symContactMapEvaluator.evaluateDifferenceDistance(contactMap, clusterContactMap)
        else:
            raise ValueError("Evaluator type %s not found!!" % self.typeEvaluator)


def getAllTrajectories(paths):
    """
        Find all the trajectory files in the paths specified

        :param paths: The path where to find the trajectories
        :type paths: str
        :returns: list -- A list with the names of all th trajectories in paths
    """
    files = []
    for path in paths:
        files += glob.glob(path)
    # sort the files obtained by glob by name, so that the results will be the
    # same on all computers
    return sorted(files)



from __future__ import absolute_import, division, print_function, unicode_literals
import unittest
import numpy as np
import AdaptivePELE.spawning.spawning as spawning
from AdaptivePELE.clustering import clustering
from AdaptivePELE.spawning import densitycalculator


class TestSpawningCalculator(unittest.TestCase):
    def testDivideTrajAccordingToWeights(self):
        spawningCalculator = spawning.SpawningCalculator()
        weights = [0.5, 0.2, 0.2, 0.1]
        trajToDistribute1 = 12
        degeneracy = spawningCalculator.divideTrajAccordingToWeights(weights, trajToDistribute1)

        golden = [6, 2, 3, 1]
        self.assertEqual(degeneracy, golden)

    def testDivideInverselyProportionalToArray(self):
        spawningCalculator = spawning.SpawningCalculator()
        weights = np.array([0.5, 0.2, 0.2, 0.1])
        trajToDistribute = 12
        degeneracy = spawningCalculator.divideInverselyProportionalToArray(weights, trajToDistribute)

        golden = [1, 3, 3, 5]

        self.assertEqual(degeneracy, golden)

    def testInverselyProportionalToPopulationCalculator(self):
        clusters = clustering.Clusters()
        sizes = [6, 2, 3, 1]
        for size in sizes:
            cluster = clustering.Cluster(None, None, None, None)
            cluster.elements = size
            clusters.addCluster(cluster)

        inverselyProp = spawning.InverselyProportionalToPopulationCalculator()
        clusteringParams = None
        trajs = 10
        degeneracy = inverselyProp.calculate(clusters.clusters, trajs, clusteringParams)
        golden = [1, 2, 2, 5]

        self.assertEqual(degeneracy, golden)

    def testInverselyProportionalToPopulationCalculatorWithNullDensity(self):
        clusters = clustering.Clusters()
        sizes = [6, 2, 3, 1]
        for size in sizes:
            cluster = clustering.Cluster(None, None, None, None)
            cluster.elements = size
            clusters.addCluster(cluster)

        densityCalculator = densitycalculator.NullDensityCalculator()

        inverselyProp = spawning.InverselyProportionalToPopulationCalculator(densityCalculator)
        clusteringParams = None
        trajs = 10
        degeneracy = inverselyProp.calculate(clusters.clusters, trajs, clusteringParams)
        golden = [1, 2, 2, 5]

        self.assertEqual(degeneracy, golden)

    def testInverselyProportionalToPopulationCalculatorWithDensity(self):
        # test 3
        clusters = clustering.Clusters()
        sizes = [6, 2, 3, 1]
        contacts = [4, 3, 2, 1]
        for size, ncontacts in zip(sizes, contacts):
            cluster = clustering.Cluster(None, None, None, None)
            cluster.elements = size
            cluster.contacts = ncontacts
            clusters.addCluster(cluster)

        clusteringBlock = {
            "density": {
                "type": "heaviside",
                "params": {
                    "values": [6, 2, 3, 1],
                    "conditions": [3, 2, 1]
                }
            }
        }
        densityCalculatorBuilder = densitycalculator.DensityCalculatorBuilder()
        densityCalculator = densityCalculatorBuilder.build(clusteringBlock)

        inverselyProp = spawning.InverselyProportionalToPopulationCalculator(densityCalculator)
        clusteringParams = None
        trajs = 8
        degeneracy = inverselyProp.calculate(clusters.clusters, trajs, clusteringParams)
        golden = [2, 2, 2, 2]

        self.assertEqual(degeneracy, golden)

    def testEpsilonCalculator(self):
        epsilon = spawning.EpsilonDegeneracyCalculator()
        params = spawning.SpawningParams()
        params.epsilon = 0.5
        params.metricWeights = "linear"
        params.nclusters = 100

        clusters = clustering.Clusters()
        sizes = [6, 2, 3, 1]
        energies = [-4, -2, -2, -1]
        for size, energy in zip(sizes, energies):
            cluster = clustering.Cluster(None, None, None, None, metricCol=0)
            cluster.elements = size
            cluster.metrics = [energy]
            clusters.addCluster(cluster)

        trajs = 20
        degeneracy = epsilon.calculate(clusters.clusters, trajs, params)
        golden = np.array([7, 4, 4, 5])
        np.testing.assert_array_equal(degeneracy, golden)

    def testEpsilonCalculatorWithDensity(self):
        params = spawning.SpawningParams()
        params.epsilon = 0.5
        params.metricWeights = "linear"
        params.nclusters = 100

        clusters = clustering.Clusters()
        sizes = [6, 2, 3, 1]
        energies = [-4, -2, -2, 0]
        contacts = [4, 3, 2, 1]
        for size, energy, ncontacts in zip(sizes, energies, contacts):
            cluster = clustering.Cluster(None, None, None, None, metricCol=0)
            cluster.elements = size
            cluster.metrics = [energy]
            cluster.contacts = ncontacts
            clusters.addCluster(cluster)

        clusteringBlock = {
            "density": {
                "type": "heaviside",
                "params": {
                    "values": [6, 2, 3, 1],
                    "conditions": [3, 2, 1]
                }
            }
        }
        densityCalculatorBuilder = densitycalculator.DensityCalculatorBuilder()
        densityCalculator = densityCalculatorBuilder.build(clusteringBlock)

        epsilon = spawning.EpsilonDegeneracyCalculator(densityCalculator)
        trajs = 16
        degeneracy = epsilon.calculate(clusters.clusters, trajs, params)
        golden = np.array([6, 4, 4, 2])
        np.testing.assert_array_equal(degeneracy, golden)

    def testEpsilonCalculatorwithMaxValues(self):
        epsilon = spawning.EpsilonDegeneracyCalculator()
        params = spawning.SpawningParams()
        params.epsilon = 0.5
        params.metricWeights = "linear"
        params.nclusters = 100
        params.condition = "max"

        clusters = clustering.Clusters()
        sizes = [6, 2, 3, 1]
        energies = [-4, -2, -2, -1]
        for size, energy in zip(sizes, energies):
            cluster = clustering.Cluster(None, None, None, None, metricCol=0)
            cluster.elements = size
            cluster.metrics = [energy]
            clusters.addCluster(cluster)

        trajs = 20
        degeneracy = epsilon.calculate(clusters.clusters, trajs, params)
        golden = np.array([1, 5, 5, 9])
        np.testing.assert_array_equal(degeneracy, golden)

    def testSameWeightDegeneracyCalculator(self):
        sameWeightDegCalculator = spawning.SameWeightDegeneracyCalculator()
        params = None

        clusters = clustering.Clusters()
        sizes = [6, 2, 3, 1]
        for size in sizes:
            cluster = clustering.Cluster(None, None, None, None)
            cluster.elements = size
            clusters.addCluster(cluster)

        trajs = 10
        degeneracy = sameWeightDegCalculator.calculate(clusters.clusters, trajs, params)
        golden = [1, 1, 1, 1]

        self.assertEqual(degeneracy, golden)

    def testVariableEpsilonCalculator(self):
        variable_epsilon = spawning.VariableEpsilonDegeneracyCalculator()
        params = spawning.SpawningParams()
        params.epsilon = 0.5
        params.varEpsilonType = "linearVariation"
        params.maxEpsilon = 0.75
        params.minEpsilon = 0.5
        params.variationWindow = 8
        params.maxEpsilonWindow = 2
        params.metricWeights = "linear"
        params_test = {}
        params.period = params_test.get("period", params.variationWindow)
        params.nclusters = 100
        params.period += np.sign(np.abs(params.variationWindow-params.period))
        rateVariation = 0.25/3
        clusters = clustering.Clusters()
        sizes = [6, 2, 3, 1]
        energies = [-4, -2, -2, -1]
        for size, energy in zip(sizes, energies):
            cluster = clustering.Cluster(None, None, None, None, metricCol=0)
            cluster.elements = size
            cluster.metrics = [energy]
            clusters.addCluster(cluster)

        trajs = 20

        degeneracy6 = variable_epsilon.calculate(clusters.clusters, trajs, params, 0)
        golden6 = np.array([7, 4, 4, 5])
        np.testing.assert_array_equal(degeneracy6, golden6)
        self.assertAlmostEqual(params.epsilon, params.minEpsilon)
        degeneracy7 = variable_epsilon.calculate(clusters.clusters, trajs, params, 1)
        # TODO: check degeneracy after next steps
        self.assertAlmostEqual(params.epsilon, params.minEpsilon+rateVariation)
        degeneracy8 = variable_epsilon.calculate(clusters.clusters, trajs, params, 2)
        self.assertAlmostEqual(params.epsilon, params.minEpsilon+2*rateVariation)
        degeneracy9 = variable_epsilon.calculate(clusters.clusters, trajs, params, 9)
        self.assertAlmostEqual(params.epsilon, params.minEpsilon)

    def testPeriodicVariableEpsilonCalculator(self):
        variable_epsilon = spawning.VariableEpsilonDegeneracyCalculator()
        params = spawning.SpawningParams()
        params.epsilon = 0.5
        params.varEpsilonType = "linearVariation"
        params.maxEpsilon = 0.75
        params.minEpsilon = 0.5
        params.variationWindow = 20
        params.maxEpsilonWindow = 1
        params.metricWeights = "linear"
        params.nclusters = 100
        params_test = {"period": 8}
        params.period = params_test.get("period", params.variationWindow)
        params.period += np.sign(np.abs(params.variationWindow-params.period))
        rateVariation = (params.maxEpsilon-params.minEpsilon)/3
        clusters = clustering.Clusters()
        sizes = [6, 2, 3, 1]
        energies = [-4, -2, -2, -1]
        for size, energy in zip(sizes, energies):
            cluster = clustering.Cluster(None, None, None, None, metricCol=0)
            cluster.elements = size
            cluster.metrics = [energy]
            clusters.addCluster(cluster)

        trajs = 20

        degeneracy6 = variable_epsilon.calculate(clusters.clusters, trajs, params, 0)
        golden6 = np.array([7, 4, 4, 5])
        np.testing.assert_array_equal(degeneracy6, golden6)
        self.assertAlmostEqual(params.epsilon, params.minEpsilon)
        for i in range(1, params.variationWindow):
            degeneracy7 = variable_epsilon.calculate(clusters.clusters, trajs, params, i)
        # TODO: check degeneracy after next steps
        # self.assertAlmostEqual(params.epsilon, params.minEpsilon+rateVariation)
        # degeneracy8 = variable_epsilon.calculate(clusters.clusters, trajs, params, 2)
        # self.assertAlmostEqual(params.epsilon, params.maxEpsilon)
        # degeneracy8 = variable_epsilon.calculate(clusters.clusters, trajs, params, 3)
        # self.assertAlmostEqual(params.epsilon,params.maxEpsilon-rateVariation/2)
        # degeneracy8 = variable_epsilon.calculate(clusters.clusters, trajs, params, 4)
        # self.assertAlmostEqual(params.epsilon, params.minEpsilon)
        # degeneracy8 = variable_epsilon.calculate(clusters.clusters, trajs, params, 5)
        # self.assertAlmostEqual(params.epsilon, params.minEpsilon)
        # degeneracy9 = variable_epsilon.calculate(clusters.clusters, trajs, params, 9)
        # self.assertAlmostEqual(params.epsilon, params.minEpsilon)

    def testUCBCalculator(self):
        UCB = spawning.UCBCalculator()
        params = spawning.SpawningParams()
        params.alpha = 8.0

        clusters = clustering.Clusters()
        sizes = [6, 2, 3, 1]
        energies = [-4, -2, -2, -1]
        for size, energy in zip(sizes, energies):
            cluster = clustering.Cluster(None, None, None, None, metricCol=0)
            cluster.elements = size
            cluster.metrics = [energy]
            clusters.addCluster(cluster)

        trajs = 20
        degeneracy = UCB.calculate(clusters.clusters, trajs, params)
        golden = np.array([3, 5, 3, 9])
        np.testing.assert_array_equal(degeneracy, golden)


def main():
    return unittest.main(exit=False)

if __name__ == '__main__':
    main()



from __future__ import absolute_import, division, print_function, unicode_literals
import sys
import shutil
import os
import json
import time
import glob
import argparse
import atexit
import ast
import numpy as np
from builtins import range
from AdaptivePELE.constants import blockNames, constants
from AdaptivePELE.atomset import atomset
from AdaptivePELE.utilities import utilities
from AdaptivePELE.validator import controlFileValidator
from AdaptivePELE.spawning import spawning, spawningTypes
from AdaptivePELE.simulation import simulationrunner, simulationTypes
from AdaptivePELE.clustering import clustering, clusteringTypes


def parseArgs():
    parser = argparse.ArgumentParser(description="Perform several iterations of"
                                     " simulations using adaptive sampling to "
                                     "distribute the processors in order "
                                     "to optimize sampling")
    parser.add_argument('controlFile', type=str)
    arg = parser.parse_args()
    return arg


class EmptyInitialStructuresError(Exception):
    __module__ = Exception.__module__


def checkMetricExitConditionMultipleTrajsinRestart(firstRun, outputFolder, simulationRunner):
    """
        Check the previous simulation data when restarting a simulation with a
        multiple trajectory metric exit condition

        :param firstRun: First epoch to be run in restart simulation
        :type firstRun: int
        :param outputFolder: Folder of the previous simulation data
        :type outputFolder: str
        :param simulationRunner: Simulation runner object
        :type simulationRunner: :py:class:`.SimulationRunner`
    """
    if simulationRunner.hasExitCondition():
        if simulationRunner.parameters.exitCondition.type != blockNames.ExitConditionType.metricMultipleTrajs:
            return
        for i in range(firstRun):
            simulationRunner.parameters.exitCondition.checkExitCondition(outputFolder % i)


def filterClustersAccordingToBox(simulationRunnerParams, clusteringObject):
    """
        Filter the clusters to select only the ones whose representative
        structures will fit into the selected box

        :param simulationRunnerParams: :py:class:`.SimulationParameters` Simulation parameters object
        :type simulationRunnerParams: :py:class:`.SimulationParameters`
        :param clusteringObject: Clustering object
        :type clusteringObject: :py:class:`.Clustering`

        :returns list, list: -- list of the filtered clusters, list of bools flagging wether the cluster is selected or not

    """
    box_center = ast.literal_eval(simulationRunnerParams.boxCenter)
    box_radius = simulationRunnerParams.boxRadius
    clustersFiltered = []
    clustersSelected = []
    for cluster in clusteringObject.clusters.clusters:
        if utilities.distanceCOM(box_center, cluster.pdb.getCOM()) < (box_radius-1):
            clustersFiltered.append(cluster)
            clustersSelected.append(True)
        else:
            clustersSelected.append(False)
    return clustersFiltered, clustersSelected


def mergeFilteredClustersAccordingToBox(degeneracy, clustersFiltering):
    """
        Merge the (possibly) partial degeneracy to obtain a complete list,
        the partial list comes from the clusters excluded by the moving box
        :param degeneracy: Degeneracy of the clusters
        :param degeneracy: list
        :param clustersFiltering: List of bools indicating whether a cluster was filtered or not
        :param clustersFiltering: list

        :returns list: -- complete list of cluster degeneracies
    """
    assert len(degeneracy) == sum(clustersFiltering)
    newDegeneracy = []
    for filtered in clustersFiltering:
        if filtered:
            newDegeneracy.append(degeneracy.pop(0))
        else:
            newDegeneracy.append(0)
    return np.array(newDegeneracy)


def expandInitialStructuresWildcard(initialStructuresWildcard):
    """
        Returns the initial structures after expanding the initial structures wildcard

        :param initialStructureWildcard: Wildcard that matches the initial structures
        :type initialStructureWildcard: str

        :return: list of str -- The expanded initial structures
    """
    totalInitialStructures = []
    for initialStructureWildcard in initialStructuresWildcard:
        expandedStructures = glob.glob(initialStructureWildcard)
        totalInitialStructures.extend(expandedStructures)
    return totalInitialStructures


def checkSymmetryDict(clusteringBlock, initialStructures, resname):
    """
        Check if the symmetries dictionary is valid for the ligand

        :param clusteringBlock: JSON block with the clustering-related parameters
        :type clusteringBlock: json
        :param initialStructures: List with initial structures
        :type initialStructures: list
        :param resname: Residue name of the ligand in the system pdb
        :type resname: str

        :raise AssertionError: If atoms are not found in the structure
     """
    symmetries = clusteringBlock[blockNames.ClusteringTypes.params].get(blockNames.ClusteringTypes.symmetries, {})
    for structure in initialStructures:
        PDB = atomset.PDB()
        PDB.initialise(structure, resname=resname)
        utilities.assertSymmetriesDict(symmetries, PDB)


def fixReportsSymmetry(outputPath, resname, nativeStructure, symmetries):
    """
        Adds a new column in the report file with the RMSD that takes into account symmetries.
        New reports are stored in the fixedReport_i where i is the number of the report

        :param outputPath: Path where trajectories are found
        :type outputPath: str
        :param resname: Residue name of the ligand in the pdb
        :type resname: str
        :param nativeStructure: Path to the native structure pdb
        :type nativeStructure: str
        :param symmetries: Dictionary containg the symmetries of the ligand
        :type symmetries: dict

        :raise IndexError: If original report file not found in output folder
    """
    outputFilename = "fixedReport_%d"  # move to constants?
    trajName = "*traj*.pdb"  # move to constants?
    reportName = "*report_%d"  # move to constants?
    trajs = glob.glob(os.path.join(outputPath, trajName))
    nativePDB = atomset.PDB()
    nativePDB.initialise(str(nativeStructure), resname=resname)
    for traj in trajs:
        trajNum = utilities.getTrajNum(traj)
        rmsd = list(utilities.getRMSD(traj, nativePDB, resname, symmetries))
        try:
            reportFilename = glob.glob(os.path.join(outputPath, reportName) % trajNum)[0]
        except IndexError:
            raise IndexError("File %s not found in folder %s" % (reportName % trajNum, outputPath))
        rmsd.insert(0, "\tCorrected RMSD")
        with open(reportFilename, "r") as f:
            report = f.readlines()
        outfile = open(os.path.join(outputPath, outputFilename % trajNum), "w")
        for line, value in zip(report, rmsd):
            outfile.write(line.rstrip("\n")+str(value)+"\n")
        outfile.close()


def copyInitialStructures(initialStructures, tmpInitialStructuresTemplate, iteration):
    """
        Copies the initial structures from a certain iteration

        :param initialStructures: Name of the initial structures to copy
        :type initialStructures: list of str
        :param tmpInitialStructuresTemplate: Template with the name of the initial strutctures
        :type tmpInitialStructuresTemplate: str
        :param iteration: Epoch number
        :type iteration: int
    """

    for i, name in enumerate(initialStructures):
        shutil.copyfile(name, tmpInitialStructuresTemplate % (iteration, i))


def generateTrajectorySelectionString(epoch, epochOutputPathTempletized):
    """
        Generates the template for the name of the trajectories in a given epoch

        :param Poch: Epoch number
        :type epoch: int
        :param epochOutputPathTempletized: Templetized path where the trajectories of any epoch are stored
        :type epochOutputPathTempletized: str

        :returns: str -- Template for the name of the trajectories in a given
            epoch
    """
    return "[\"" + os.path.join(epochOutputPathTempletized % epoch, constants.trajectoryBasename) + "\"]"


def findFirstRun(outputPath, clusteringOutputObject):
    """
        Find the last epoch that was properly simulated and clusterized and
        and return the first epoch to run in case of restart

        :param outputPath: Simulation output path
        :type outputPath: str
        :param clusteringOutputObject: Templetized name of the clustering object
        :type clusteringOutputObject: str

        :return: int -- Current epoch
    """

    folderWithSimulationData = outputPath
    allFolders = os.listdir(folderWithSimulationData)
    epochFolders = [int(epoch) for epoch in allFolders if epoch.isdigit()]
    epochFolders.sort(reverse=True)

    objectsFound = []
    for epoch in epochFolders:
        if os.path.exists(clusteringOutputObject % epoch):
            objectsFound.append(epoch)
        if objectsFound and epoch < (objectsFound[0]-5):
            break
    while objectsFound:
        epoch = objectsFound.pop(0)
        if checkIntegrityClusteringObject(clusteringOutputObject % epoch):
            return epoch + 1
    return 0


def checkIntegrityClusteringObject(objectPath):
    """
        Test whether the found clustering object to reload is a valid object

        :param objectPath: Clustering object path
        :type objectPath: str

        :returns: bool -- True if the found clustering object is valid
    """
    try:
        utilities.readClusteringObject(objectPath)
        return True
    except EOFError:
        return False


def __unicodeToStr(data):
    # convert dict
    if isinstance(data, dict):
        return {__unicodeToStr(key): __unicodeToStr(value) for key, value in data.items()}
    # convert list
    if isinstance(data, list):
        return [__unicodeToStr(val) for val in data]
    # convert unicode to str
    if isinstance(data, unicode):
        return data.encode('utf-8')

    return data


def loadParams(jsonParams):
    """
        Read the control file in JSON format and extract the blocks of simulation,
        general parameters, clustering and spawning

        :param jsonParams: Control file in JSON format from where the parameters will be read
        :type jsonParams: json str
    """
    jsonFile = open(jsonParams, 'r').read()
    # parsedJSON = json.loads(jsonFile, object_hook=__unicodeToStr)
    parsedJSON = json.loads(jsonFile)

    return parsedJSON[blockNames.ControlFileParams.generalParams], parsedJSON[blockNames.ControlFileParams.spawningBlockname],\
        parsedJSON[blockNames.ControlFileParams.simulationBlockname], parsedJSON[blockNames.ControlFileParams.clusteringBlockname]


def saveInitialControlFile(jsonParams, originalControlFile):
    """
        Save the adaptive control file jsonParams in originalControlFile

        :param jsonParams: Input control file in JSON format
        :type jsonParams: str
        :param originalControlFile: Path where to save the control file
        :type originalControlFile: str
    """
    with open(originalControlFile, 'w') as f:
        with open(jsonParams, 'r') as fr:
            jsonFile = fr.read()
        f.write(jsonFile)


def needToRecluster(oldClusteringMethod, newClusteringMethod):
    """
        Check if the parameters have changed in a restart and we need to redo
        the clustering. In particular: type of clustering, theshold calculator
        or distance

        :param oldClusteringMethod: Clustering in a previous simulation before the restart
        :type oldClusteringMethod: :py:class:`.Clustering`
        :param newClusteringMethod: Clustering in the restarted simulation
        :type newClusteringMethod: :py:class:`.Clustering`

        :returns: bool -- If clustering needs to be redone
    """

    # Check 1: change of type
    if oldClusteringMethod.type != newClusteringMethod.type:
        return True

    # Check 2: Change of thresholdCalculator and thresholdDistance
    if oldClusteringMethod.type == clusteringTypes.CLUSTERING_TYPES.rmsd or\
       oldClusteringMethod.type == clusteringTypes.CLUSTERING_TYPES.contactMap:
        return oldClusteringMethod.thresholdCalculator != newClusteringMethod.thresholdCalculator or\
                abs(oldClusteringMethod.contactThresholdDistance - newClusteringMethod.contactThresholdDistance) > 1e-7

    # Check 3: Change of similarity Evaluator in contactMap clustering
    if oldClusteringMethod.type == clusteringTypes.CLUSTERING_TYPES.contactMap:
        return oldClusteringMethod.similarityEvaluator.typeEvaluator != newClusteringMethod.similarityEvaluator.typeEvaluator


def clusterEpochTrajs(clusteringMethod, epoch, epochOutputPathTempletized):
    """
        Cluster the trajecotories of a given epoch

        :param clusteringMethod: Clustering object
        :type clusteringMethod: :py:class:`.Clustering`
        :param epoch: Number of the epoch to cluster
        :type epoch: int
        :param epochOutputPathTempletized: Path where to find the trajectories
        :type epochOutputPathTempletized: str
"""

    snapshotsJSONSelectionString = generateTrajectorySelectionString(epoch, epochOutputPathTempletized)
    paths = ast.literal_eval(snapshotsJSONSelectionString)
    if len(glob.glob(paths[-1])) == 0:
        sys.exit("No trajectories to cluster! Matching path:%s" % paths[-1])
    clusteringMethod.cluster(paths)


def clusterPreviousEpochs(clusteringMethod, finalEpoch, epochOutputPathTempletized, simulationRunner):
    """
        Cluster all previous epochs using the clusteringMethod object

        :param clusteringMethod: Clustering object
        :type clusteringMethod: :py:class:`.Clustering`
        :param finalEpoch: Last epoch to cluster (not included)
        :type finalEpoch: int
        :param epochOutputPathTempletized: Path where to find the trajectories
        :type epochOutputPathTempletized: str
        :param simulationRunner: Simulation runner object
        :type simulationRunner: :py:class:`.SimulationRunner`
"""
    for i in range(finalEpoch):
        simulationRunner.readMappingFromDisk(epochOutputPathTempletized % i)
        clusterEpochTrajs(clusteringMethod, i, epochOutputPathTempletized)


def getWorkingClusteringObjectAndReclusterIfNecessary(firstRun, outputPathConstants, clusteringBlock, spawningParams, simulationRunner):
    """
        It reads the previous clustering method, and, if there are changes,
        it reclusters the previous trajectories. Returns the clustering object to use

        :param firstRun: New epoch to run
        :type firstRun: int
        :param outputPathConstants: Contains outputPath-related constants
        :type outputPathConstants: :py:class:`.OutputPathConstants`
        :param clusteringBlock: Contains the new clustering block
        :type clusteringBlock: json
        :param spawningParams: Spawning params, to know what reportFile and column to read
        :type spawningParams: :py:class:`.SpawningParams`

        :returns: :py:class:`.Clustering` -- The clustering method to use in the
            adaptive sampling simulation
    """

    lastClusteringEpoch = firstRun - 1
    clusteringObjectPath = outputPathConstants.clusteringOutputObject % (lastClusteringEpoch)
    oldClusteringMethod = utilities.readClusteringObject(clusteringObjectPath)

    clusteringBuilder = clustering.ClusteringBuilder()
    clusteringMethod = clusteringBuilder.buildClustering(clusteringBlock,
                                                         spawningParams.reportFilename,
                                                         spawningParams.reportCol)

    if needToRecluster(oldClusteringMethod, clusteringMethod):
        print("Reclustering!")
        startTime = time.time()
        clusterPreviousEpochs(clusteringMethod, firstRun, outputPathConstants.epochOutputPathTempletized, simulationRunner)
        endTime = time.time()
        print("Reclustering took %s sec" % (endTime - startTime))
    else:
        clusteringMethod = oldClusteringMethod
        clusteringMethod.setCol(spawningParams.reportCol)

    return clusteringMethod


def buildNewClusteringAndWriteInitialStructuresInRestart(firstRun, outputPathConstants, clusteringBlock,
                                                         spawningParams, spawningCalculator, simulationRunner):
    """
        It reads the previous clustering method, and if there are changes (clustering method or related to thresholds),
        reclusters the previous trajectories. Returns the clustering object to use,
        and the initial structure filenames as strings

        :param firstRun: New epoch to run
        :type firstRun: int
        :param outputPathConstants: Contains outputPath-related constants
        :type outputPathConstants: str
        :param clusteringBlock: Contains the new clustering block
        :type clusteringBlock: json
        :param spawningParams: Spawning params
        :type spawningParams: :py:class:`.SpawningParams`
        :param spawningCalculator: Spawning calculator object
        :type spawningCalculator: :py:class:`.SpawningCalculator`
        :param simulationRunner: :py:class:`.SimulationRunner` Simulation runner object
        :type simulationRunner: :py:class:`.SimulationRunner`

        :returns: :py:class:`.Clustering`, str -- The clustering method to use in the adaptive sampling simulation and the initial structures filenames
    """

    clusteringMethod = getWorkingClusteringObjectAndReclusterIfNecessary(firstRun, outputPathConstants, clusteringBlock, spawningParams, simulationRunner)

    degeneracyOfRepresentatives = spawningCalculator.calculate(clusteringMethod.clusters.clusters, simulationRunner.parameters.processors-1, spawningParams, firstRun)
    spawningCalculator.log()
    print("Degeneracy", degeneracyOfRepresentatives)
    seedingPoints, procMapping = spawningCalculator.writeSpawningInitialStructures(outputPathConstants, degeneracyOfRepresentatives, clusteringMethod, firstRun)
    initialStructuresAsString = simulationRunner.createMultipleComplexesFilenames(seedingPoints, outputPathConstants.tmpInitialStructuresTemplate, firstRun)
    simulationRunner.updateMappingProcessors(procMapping)

    return clusteringMethod, initialStructuresAsString


def buildNewClusteringAndWriteInitialStructuresInNewSimulation(debug, controlFile, outputPathConstants, clusteringBlock, spawningParams, initialStructures, simulationRunner):
    """
        Build the clustering object and copies initial structures from control file.
        Returns the clustering object to use and the initial structures filenames as string

        :param debug: In debug, it will not remove the simulations
        :type debug: bool
        :param controlFile: Adaptive sampling control file
        :type controlFile: str
        :param outputPathConstants: Contains outputPath-related constants
        :type outputPathConstants: :py:class:`.OutputPathConstants`
        :param clusteringBlock: Contains the new clustering block
        :type clusteringBlock: json
        :param spawningParams: Spawning params
        :type spawningParams: :py:class:`.SpawningParams`
        :param initialStructures: Control file initial structures
        :type initialStructures: list
        :param simulationRunner: :py:class:`.SimulationRunner` Simulation runner object
        :type simulationRunner: :py:class:`.SimulationRunner`

        :returns: :py:class:`.Clustering`, str -- The clustering method to use in the adaptive sampling simulation and the initial structures filenames
    """
    saveInitialControlFile(controlFile, outputPathConstants.originalControlFile)

    firstRun = 0
    copyInitialStructures(initialStructures, outputPathConstants.tmpInitialStructuresTemplate, firstRun)
    initialStructuresAsString = simulationRunner.createMultipleComplexesFilenames(len(initialStructures), outputPathConstants.tmpInitialStructuresTemplate, firstRun)

    clusteringBuilder = clustering.ClusteringBuilder()
    clusteringMethod = clusteringBuilder.buildClustering(clusteringBlock,
                                                         spawningParams.reportFilename,
                                                         spawningParams.reportCol)
    initialClusters = []
    return clusteringMethod, initialStructuresAsString, initialClusters


def preparePeleControlFile(epoch, outputPathConstants, simulationRunner, peleControlFileDictionary):
    """
        Substitute the parameters in the PELE control file specified with the
        provided in the control file

        :param epoch: Epoch number
        :type epoch: int
        :param outputPathConstants: Object that has as attributes constant related to the outputPath that will be used to create the working control file
        :type outputPathConstants: :py:class:`.OutputPathConstants`
        :param simulationRunner: Simulation runner object
        :type simulationRunner: `.SimulationRunner`
        :param peleControlFileDictionary: Dictonary containing the values of the parameters to substitute in the control file
        :type peleControlFileDictionary: dict
    """
    outputDir = outputPathConstants.epochOutputPathTempletized % epoch
    utilities.makeFolder(outputDir)
    peleControlFileDictionary["OUTPUT_PATH"] = outputDir
    peleControlFileDictionary["SEED"] = simulationRunner.parameters.seed + epoch*simulationRunner.parameters.processors
    if simulationRunner.parameters.boxCenter is not None:
        peleControlFileDictionary["BOX_RADIUS"] = simulationRunner.parameters.boxRadius
        peleControlFileDictionary["BOX_CENTER"] = simulationRunner.parameters.boxCenter
    simulationRunner.makeWorkingControlFile(outputPathConstants.tmpControlFilename % epoch, peleControlFileDictionary)


def main(jsonParams, clusteringHook=None):
    """
        Main body of the adaptive sampling program.

        :param jsonParams: A string with the name of the control file to use
        :type jsonParams: str
    """

    controlFileValidator.validate(jsonParams)
    generalParams, spawningBlock, simulationrunnerBlock, clusteringBlock = loadParams(jsonParams)

    spawningAlgorithmBuilder = spawning.SpawningAlgorithmBuilder()
    spawningCalculator, spawningParams = spawningAlgorithmBuilder.build(spawningBlock)

    runnerbuilder = simulationrunner.RunnerBuilder()
    simulationRunner = runnerbuilder.build(simulationrunnerBlock)

    restart = generalParams.get(blockNames.GeneralParams.restart, True)
    debug = generalParams.get(blockNames.GeneralParams.debug, False)
    outputPath = generalParams[blockNames.GeneralParams.outputPath]
    initialStructuresWildcard = generalParams[blockNames.GeneralParams.initialStructures]
    writeAll = generalParams.get(blockNames.GeneralParams.writeAllClustering, False)
    nativeStructure = generalParams.get(blockNames.GeneralParams.nativeStructure, '')
    resname = str(clusteringBlock[blockNames.ClusteringTypes.params][blockNames.ClusteringTypes.ligandResname])

    print("================================")
    print("            PARAMS              ")
    print("================================")
    print("Restarting simulations", restart)
    print("Debug:", debug)

    print("Iterations: %d, Mpi processors: %d, Pele steps: %d" % (simulationRunner.parameters.iterations, simulationRunner.parameters.processors, simulationRunner.parameters.peleSteps))

    print("SpawningType:", spawningTypes.SPAWNING_TYPE_TO_STRING_DICTIONARY[spawningCalculator.type])

    print("SimulationType:", simulationTypes.SIMULATION_TYPE_TO_STRING_DICTIONARY[simulationRunner.type])
    if simulationRunner.hasExitCondition():
        print("Exit condition:", simulationTypes.EXITCONDITION_TYPE_TO_STRING_DICTIONARY[simulationRunner.parameters.exitCondition.type])
    print("Clustering method:", clusteringBlock[blockNames.ClusteringTypes.type])

    print("Output path: ", outputPath)
    print("Initial Structures: ", initialStructuresWildcard)
    print("================================\n\n")

    initialStructures = expandInitialStructuresWildcard(initialStructuresWildcard)
    if not initialStructures:
        raise EmptyInitialStructuresError("No initial structures found!!!")
    checkSymmetryDict(clusteringBlock, initialStructures, resname)

    outputPathConstants = constants.OutputPathConstants(outputPath)

    if not debug:
        atexit.register(utilities.cleanup, outputPathConstants.tmpFolder)

    utilities.makeFolder(outputPath)
    utilities.makeFolder(outputPathConstants.tmpFolder)
    saveInitialControlFile(jsonParams, outputPathConstants.originalControlFile)

    startFromScratch = False
    if restart:
        firstRun = findFirstRun(outputPath, outputPathConstants.clusteringOutputObject)
        if firstRun == 0:
            startFromScratch = True
        else:
            clusteringMethod, initialStructuresAsString = buildNewClusteringAndWriteInitialStructuresInRestart(firstRun, outputPathConstants, clusteringBlock, spawningParams, spawningCalculator, simulationRunner)
            checkMetricExitConditionMultipleTrajsinRestart(firstRun, outputPathConstants.epochOutputPathTempletized, simulationRunner)

    if startFromScratch or not restart:
        firstRun = 0  # if restart false, but there were previous simulations
        # These two lines were previously inside the
        # buildNewClusteringAndWriteInitialStructuresInNewSimulation function,
        # however, the equilibration procedure has to be called in between
        if not debug:
            shutil.rmtree(outputPath)
        utilities.makeFolder(outputPath)

        if simulationRunner.parameters.runEquilibration:
            initialStructures = simulationRunner.equilibrate(initialStructures, outputPathConstants, spawningParams.reportFilename, outputPath, resname)
        clusteringMethod, initialStructuresAsString, _ = buildNewClusteringAndWriteInitialStructuresInNewSimulation(debug, jsonParams, outputPathConstants, clusteringBlock, spawningParams, initialStructures, simulationRunner)
    peleControlFileDictionary = {"COMPLEXES": initialStructuresAsString, "PELE_STEPS": simulationRunner.parameters.peleSteps,
                                 "BOX_RADIUS": simulationRunner.parameters.boxRadius}
    if simulationRunner.parameters.modeMovingBox is not None and simulationRunner.parameters.boxCenter is None:
        simulationRunner.parameters.boxCenter = simulationRunner.selectInitialBoxCenter(initialStructuresAsString, resname)

    for i in range(firstRun, simulationRunner.parameters.iterations):
        print("Iteration", i)

        print("Preparing control file...")
        preparePeleControlFile(i, outputPathConstants, simulationRunner, peleControlFileDictionary)

        print("Production run...")
        if not debug:
            startTime = time.time()
            simulationRunner.runSimulation(outputPathConstants.tmpControlFilename % i)
            endTime = time.time()
            print("PELE %s sec" % (endTime - startTime))

        simulationRunner.writeMappingToDisk(outputPathConstants.epochOutputPathTempletized % i)

        print("Clustering...")
        startTime = time.time()
        clusterEpochTrajs(clusteringMethod, i, outputPathConstants.epochOutputPathTempletized)
        endTime = time.time()
        print("Clustering ligand: %s sec" % (endTime - startTime))

        if clusteringHook is not None:
            clusteringHook(clusteringMethod, outputPathConstants, simulationRunner, i+1)

        if simulationRunner.parameters.modeMovingBox is not None:
            simulationRunner.getNextIterationBox(clusteringMethod, outputPathConstants.epochOutputPathTempletized % i, resname)
            clustersList, clustersFiltered = filterClustersAccordingToBox(simulationRunner.parameters, clusteringMethod)
        else:
            clustersList = clusteringMethod.clusters.clusters

        degeneracyOfRepresentatives = spawningCalculator.calculate(clustersList, simulationRunner.parameters.processors-1, spawningParams, i)
        spawningCalculator.log()

        if degeneracyOfRepresentatives is not None:
            if simulationRunner.parameters.modeMovingBox is not None:
                degeneracyOfRepresentatives = mergeFilteredClustersAccordingToBox(degeneracyOfRepresentatives, clustersFiltered)
            print("Degeneracy", degeneracyOfRepresentatives)
            assert len(degeneracyOfRepresentatives) == len(clusteringMethod.clusters.clusters)
        else:
            # When using null spawning the calculate method returns None
            assert spawningCalculator.type == spawningTypes.SPAWNING_TYPES.null, "calculate returned None with spawning type %s" % spawningTypes.SPAWNING_TYPE_TO_STRING_DICTIONARY[spawningCalculator.type]

        clusteringMethod.writeOutput(outputPathConstants.clusteringOutputDir % i,
                                     degeneracyOfRepresentatives,
                                     outputPathConstants.clusteringOutputObject % i, writeAll)

        if i > 0:
            # Remove old clustering object, since we already have a newer one
            try:
                os.remove(outputPathConstants.clusteringOutputObject % (i-1))
            except OSError:
                # In case of restart
                pass

        # Prepare for next pele iteration
        if i != simulationRunner.parameters.iterations-1:
            if degeneracyOfRepresentatives is not None:
                numberOfSeedingPoints, procMapping = spawningCalculator.writeSpawningInitialStructures(outputPathConstants, degeneracyOfRepresentatives, clusteringMethod, i+1)
                simulationRunner.updateMappingProcessors(procMapping)
                initialStructuresAsString = simulationRunner.createMultipleComplexesFilenames(numberOfSeedingPoints, outputPathConstants.tmpInitialStructuresTemplate, i+1)
                peleControlFileDictionary["COMPLEXES"] = initialStructuresAsString

        if clusteringMethod.symmetries and nativeStructure:
            fixReportsSymmetry(outputPathConstants.epochOutputPathTempletized % i, resname,
                               nativeStructure, clusteringMethod.symmetries)

        # check exit condition, if defined
        if simulationRunner.hasExitCondition():
            if simulationRunner.checkExitCondition(clusteringMethod, outputPathConstants.epochOutputPathTempletized % i):
                print("Simulation exit condition met at iteration %d, stopping" % i)
                break
            else:
                print("Simulation exit condition not met at iteration %d, continuing..." % i)

if __name__ == '__main__':
    args = parseArgs()
    main(args.controlFile)



from __future__ import absolute_import, division, print_function, unicode_literals
from AdaptivePELE.clustering import clustering
import unittest


class clusteringTest(unittest.TestCase):
    def testCluster(self):
        # preparation
        clusteringBuilder = clustering.ClusteringBuilder()
        clusteringParams = {"type": "rmsd",
                            "params": {"ligandResname": "AIN",
                                       "contactThresholdDistance": 8}}
        clusteringInstance = clusteringBuilder.buildClustering(clusteringParams,
                                                               "ain_report", 3)

        trajNames = ["tests/data/aspirin_data/traj*"]

        # function to test
        clusteringInstance.cluster(trajNames)

        # assertion
        allClusters = clusteringInstance.clusters.clusters
        goldenNumberOfClusters = 2
        goldenEnergyCluster1 = -8424.8
        goldenEnergyCluster2 = -8453.29
        goldenElementsCluster1 = 2
        goldenElementsCluster2 = 1

        self.assertEqual(len(allClusters), goldenNumberOfClusters)
        self.assertAlmostEqual(allClusters[0].getMetric(), goldenEnergyCluster1, 2)
        self.assertAlmostEqual(allClusters[1].getMetric(), goldenEnergyCluster2, 2)
        self.assertEqual(allClusters[0].elements, goldenElementsCluster1)
        self.assertEqual(allClusters[1].elements, goldenElementsCluster2)

    def test_cluster_accumulative(self):
        # preparation
        clusteringParams = {"type": "contactMap",
                            "params": {"ligandResname": "AIN",
                                       "contactThresholdDistance": 8,
                                       "similarityEvaluator": "correlation"},
                            "thresholdCalculator": {
                                "type": "constant",
                                "params": {
                                    "value": 0.15
                                }
                            }}
        clusteringBuilder = clustering.ClusteringBuilder()
        clusteringInstance = clusteringBuilder.buildClustering(clusteringParams,
                                                               "ain_report", 3)

        trajNames = ["tests/data/aspirin_data/traj*"]

        # function to test
        clusteringInstance.cluster(trajNames)
        # assertion
        allClusters = clusteringInstance.clusters.clusters
        # goldenNumberOfClusters = 3
        # goldenEnergyCluster1 = -8421.5
        # goldenEnergyCluster2 = -8424.8
        # goldenEnergyCluster3 = -8453.29
        # goldenElementsCluster1 = 1
        # goldenElementsCluster2 = 1
        # goldenElementsCluster3 = 1

        goldenNumberOfClusters = 2
        goldenEnergyCluster1 = -8424.8
        goldenEnergyCluster2 = -8453.29
        goldenElementsCluster1 = 2
        goldenElementsCluster2 = 1

        self.assertEqual(len(allClusters), goldenNumberOfClusters)
        self.assertAlmostEqual(allClusters[0].getMetric(), goldenEnergyCluster1, 2)
        self.assertAlmostEqual(allClusters[1].getMetric(), goldenEnergyCluster2, 2)
        # self.assertAlmostEqual(allClusters[2].getMetric(), goldenEnergyCluster3, 2)
        self.assertEqual(allClusters[0].elements, goldenElementsCluster1)
        self.assertEqual(allClusters[1].elements, goldenElementsCluster2)
        # self.assertEqual(allClusters[2].elements, goldenElementsCluster3)

    def test_CMInnerLimit(self):
        CMEvaluator = clustering.CMClusteringEvaluator(None, None)
        cluster1_8 = clustering.Cluster(None, contacts=1, contactThreshold=8)
        cluster2_8 = clustering.Cluster(None, contacts=2.2, contactThreshold=8)
        cluster2_6 = clustering.Cluster(None, contacts=2.2, contactThreshold=6)
        cluster1_6 = clustering.Cluster(None, contacts=1.0, contactThreshold=6)
        cluster1_4 = clustering.Cluster(None, contacts=1.0, contactThreshold=4)
        cluster2_4 = clustering.Cluster(None, contacts=2.2, contactThreshold=4)
        # Reduced slope to 8
        self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster1_8), 12)
        self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster2_8), 4)
        self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster1_6), 12)
        self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster2_6), 4)
        self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster1_4), 12)
        self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster2_4), 4)
        # Slope 14
        # self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster1_8), 18)
        # self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster2_8), 4)
        # self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster1_6), 18)
        # self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster2_6), 4)
        # self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster1_4), 18)
        # self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster2_4), 4)
        # Now the limits are always calculated with threshold 8
        # self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster1_6), 10)
        # self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster2_6), 4)
        # self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster1_4), 10)
        # self.assertAlmostEqual(CMEvaluator.getInnerLimit(cluster2_4), 4)



import sys
import os
import socket

DIR = os.path.dirname(__file__)

##############################################

# PUBLIC CONSTANTS (to change by the user)
# Preparation inputs to grow

machine = socket.getfqdn()
print(machine)
if "bsc.mn" in machine:
    # PELE parameters
    PATH_TO_PELE = "/gpfs/projects/bsc72/PELE++/mniv/rev12536/bin/Pele_mpi"
    PATH_TO_PELE_DATA = "/gpfs/projects/bsc72/PELE++/data/rev12360/Data"
    PATH_TO_PELE_DOCUMENTS = "/gpfs/projects/bsc72/PELE++/Documents/rev12360"
    PATH_TO_LICENSE = "/gpfs/projects/bsc72/PELE++/license"
    # PlopRotTemp parameters
    SCHRODINGER_PY_PATH = "/gpfs/projects/bsc72/SCHRODINGER_ACADEMIC/utilities/python"
    ENV_PYTHON = "/gpfs/projects/bsc72/SCHRODINGER_ACADEMIC/internal/lib/python2.7/site-packages/"
elif "bsccv" in machine:
    # PELE parameters
    PATH_TO_PELE = "/data/EAPM/PELE/PELE++/bin/rev12360/Pele_rev12360_mpi"
    PATH_TO_PELE_DATA = "/data/EAPM/PELE/PELE++/data/rev12360/Data"
    PATH_TO_PELE_DOCUMENTS = "/data/EAPM/PELE/PELE++/Documents/rev12360"
    PATH_TO_LICENSE = "/data/EAPM/PELE/PELE++/license"
    SCHRODINGER_PY_PATH = "/data2/bsc72/SCHRODINGER_ACADEMIC/utilities/python"
    ENV_PYTHON = "/data2/bsc72/SCHRODINGER_ACADEMIC/internal/lib/python2.7/site-packages/"
else:
    # PELE parameters
    PATH_TO_PELE = "/opt/PELE/bin/PELE-1.5"
    PATH_TO_PELE_DATA = "/opt/PELE/Data"
    PATH_TO_PELE_DOCUMENTS = "/opt/PELE/Documents"
    PATH_TO_LICENSE = "/home/ywest/"
    SCHRODINGER_PY_PATH = "/opt/schrodinger2018-1/utilities/python"
    ENV_PYTHON = ""

CONTROL_TEMPLATE = os.path.join(DIR, "Templates/control_template.conf")
RESULTS_FOLDER = "growing_output"
GROWING_STEPS = 10
SELECTION_CRITERIA = "Binding Energy"
SERIE_FILE = False
REPORT_NAME = "report"
TRAJECTORY_NAME = "trajectory"
CPUS = 48
PELE_EQ_STEPS = 20
RESTART = False
STEPS = 6
TEMPERATURE = 1000
MAX_OVERLAP = 0.70
MIN_OVERLAP = 0.50
SEED = 1279183
ROTRES = "10.0"
# Clustering parameters
DISTANCE_COUNTER = 4
CONTACT_THRESHOLD = 0.3
EPSILON = 0.5
BANNED_DIHEDRALS_ATOMS = None
BANNED_ANGLE_THRESHOLD = None
##############################################

# PRIVATE CONSTANTS (not to change)
PRE_WORKING_DIR = "pregrow"
TEMPLATES_PATH = "DataLocal/Templates/OPLS2005/HeteroAtoms/"
ROTAMERS_PATH = "DataLocal/LigandRotamerLibs/"
PDBS_OUTPUT_FOLDER = "PDBs_growing"
OUTPUT_FOLDER = "growing_results/"
TEMPLATES_FOLDER = "growing_templates"
CONFIG_PATH = "log_configure.ini"
PLOP_PATH = "PlopRotTemp_S_2017/ligand_prep.py"
# Clustering constants
CONDITION = "min"   #   min or max
METRICS_WEIGHTS = "linear"
NUM_CLUSTERS = 5
# Messages constants
TEMPLATE_MESSAGE = "We are going to transform the template _{}_ into _{}_ in _{}_ steps! Starting..."
LINES_MESSAGE = "\n•*´¨`*•.¸¸.•*´¨`*•.¸¸.•*´¨`*•.¸¸.•*´¨`*•.¸¸.••*´¨`*•.¸¸.•*´¨`*•.¸¸.•*´¨`*•.¸¸.•*´¨`*•.¸¸.•\n"
SELECTED_MESSAGE = "\n============ Files selected ============\nControl file: {}\nPDB file: {}\nResults folder name: {}\nStep: {}\n"
FINISH_SIM_MESSAGE = "SIMULATION {} COMPLETED!!! "

##############################################
